from typing import Optional

from django.utils.translation import gettext as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.operators import default, lte
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class EntryModelDisplay(DisplayViewConfig):
    def get_instance_display(self) -> Display:
        return create_simple_display([["computed_str"]])


class PersonModelDisplay(EntryModelDisplay):
    @classmethod
    def _get_person_instance_display(cls) -> Display:
        """Returns a person instance's display

        Returns:
            Display: The display instance
        """

        employers_section = Section(
            key="employers_section",
            collapsible=False,
            title=_("Employers"),
            display=Display(
                pages=[
                    Page(
                        title=_("Employers"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["employers_table"]],
                                inlines=[Inline(key="employers_table", endpoint="employers")],
                            )
                        },
                    ),
                ]
            ),
        )

        return Display(
            pages=[
                Page(
                    title=_("Main Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["profile_image", "prefix", "specializations", "activity_table"],
                                ["profile_image", "first_name", "last_name", "activity_table"],
                                ["profile_image", "primary_email", "primary_telephone", "activity_table"],
                                ["profile_image", "activity_heat", "activity_heat", "activity_table"],
                                [
                                    "personality_profile_red",
                                    "personality_profile_green",
                                    "personality_profile_blue",
                                    "activity_table",
                                ],
                                ["employers_section", "employers_section", "employers_section", "employers_section"],
                            ],
                            grid_template_columns=[
                                Style.MIN_CONTENT,
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                                "50%",
                            ],
                            grid_template_rows=[Style.rem(6), Style.rem(6), Style.rem(6), Style.rem(6)],
                            grid_auto_rows=Style.MIN_CONTENT,
                            sections=[employers_section],
                            inlines=[
                                Inline(
                                    key="activity_table",
                                    endpoint="wbcrm_activity",
                                    title="Activities",
                                    hide_controls=True,
                                ),
                            ],
                        ),
                        lte(1100): Layout(
                            grid_template_areas=[
                                ["profile_image", "prefix", "specializations"],
                                ["profile_image", "first_name", "last_name"],
                                ["profile_image", "primary_email", "primary_telephone"],
                                ["activity_heat", "activity_heat", "activity_heat"],
                                ["personality_profile_red", "personality_profile_green", "personality_profile_blue"],
                                ["employers_section", "employers_section", "employers_section"],
                            ],
                            grid_template_columns=[
                                Style.MIN_CONTENT,
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                            ],
                            grid_auto_rows=Style.MIN_CONTENT,
                            sections=[employers_section],
                        ),
                        lte(700): Layout(
                            grid_template_areas=[
                                ["first_name"],
                                ["last_name"],
                                ["primary_email"],
                                ["primary_telephone"],
                                ["activity_heat"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
                Page(
                    title=_("Additional Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["salutation", "birthday", "formal", "."],
                                ["has_user_account", "last_connection", ".", "."],
                                ["description", "description", "description", "description"],
                            ],
                            grid_template_columns=[
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                                "35%",
                            ],
                            grid_auto_rows=Style.MIN_CONTENT,
                            row_gap=Style.rem(5),
                            column_gap=Style.rem(4),
                        ),
                        lte(715): Layout(
                            grid_template_areas=[
                                ["salutation", "formal"],
                                ["birthday", "."],
                                ["has_user_account", "last_connection"],
                            ],
                            grid_template_columns=["minmax(min-content, 4fr)", "minmax(min-content, 1fr)"],
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                        lte(550): Layout(
                            grid_template_areas=[
                                ["salutation"],
                                ["birthday"],
                                ["formal"],
                                ["has_user_account"],
                                ["last_connection"],
                            ],
                            grid_auto_columns=Style.MIN_CONTENT,
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ],
        )

    @classmethod
    def _get_new_person_instance_display(cls) -> Display:
        """Returns the display for creating a new person

        Returns:
            Display: The display instance
        """

        return Display(
            pages=[
                Page(
                    title=_("Main Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["profile_image", "prefix", "specializations", "."],
                                ["profile_image", "first_name", "last_name", "."],
                                ["profile_image", "primary_email", "primary_telephone", "."],
                                ["profile_image", "primary_employer", "position_in_company", "."],
                                [".", "primary_manager", "position_name", "."],
                                [
                                    "personality_profile_red",
                                    "personality_profile_green",
                                    "personality_profile_blue",
                                    ".",
                                ],
                            ],
                            grid_template_columns=[
                                Style.MIN_CONTENT,
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                                Style.fr(1),
                            ],
                            column_gap=Style.rem(4),
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                        lte(800): Layout(
                            grid_template_areas=[
                                ["profile_image", "prefix"],
                                ["profile_image", "first_name"],
                                ["profile_image", "last_name"],
                                ["profile_image", "specializations"],
                                ["primary_email", "primary_telephone"],
                                ["primary_employer", "position_in_company"],
                                ["primary_manager", "position_name"],
                                ["personality_profile_red", "personality_profile_green"],
                                ["personality_profile_blue", "."],
                            ],
                            grid_template_columns=[Style.MIN_CONTENT, "minmax(min-content, 1fr)"],
                            column_gap=Style.rem(2),
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                        lte(450): Layout(
                            grid_template_areas=[
                                ["prefix"],
                                ["first_name"],
                                ["last_name"],
                                ["specializations"],
                                ["primary_email"],
                                ["primary_telephone"],
                                ["primary_employer"],
                                ["position_in_company"],
                                ["position_name"],
                                ["primary_manager"],
                                ["personality_profile_red"],
                                ["personality_profile_green"],
                                ["personality_profile_blue"],
                                ["prefix"],
                                ["first_name"],
                                ["last_name"],
                                ["specializations"],
                                ["primary_email"],
                                ["primary_telephone"],
                                ["primary_manager"],
                                ["personality_profile_red"],
                                ["personality_profile_green"],
                                ["personality_profile_blue"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
                Page(
                    title=_("Additional Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["salutation", "birthday", "formal", "."],
                            ],
                            grid_template_columns=[
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                                "25%",
                            ],
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(650): Layout(
                            grid_template_areas=[
                                ["birthday", "birthday"],
                                ["salutation", "formal"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                        lte(370): Layout(
                            grid_template_areas=[
                                ["salutation"],
                                ["birthday"],
                                ["formal"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ],
        )

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="primary_employer_repr", label=_("Primary Employer")),
                dp.Field(key="customer_status", label=_("Status")),
                dp.Field(key="position_in_company", label=_("Position")),
                dp.Field(key="cities", label=_("City")),
                dp.Field(key="primary_telephone", label=_("Primary Phone Number")),
                dp.Field(key="tier", label=_("Tier")),
                dp.Field(key="primary_manager_repr", label=_("Primary Relationship Manager")),
                dp.Field(
                    key=None,
                    label=_("Last Event"),
                    children=[
                        dp.Field(key="last_event", label=_("Name")),
                        dp.Field(key="last_event_period_endswith", label=_("End Date")),
                    ],
                ),
                dp.Field(key="activity_heat", label=_("Activity Heat")),
            ],
        )

    AUM_TABLE: Section = []
    PORTFOLIO_FIELDS: Section = []

    def get_instance_display(self) -> Display:
        return (
            self._get_person_instance_display()
            if "pk" in self.view.kwargs
            else self._get_new_person_instance_display()
        )


class CompanyModelDisplay(EntryModelDisplay):
    @classmethod
    def _get_company_instance_display(cls, aum_table: Section, portfolio_fields: Section) -> Display:
        """Returns a company instance's display

        Args:
            aum_table: A section with an AUM by product inline table that will be imported when wbportfolio is installed. Can be empty.
            portfolio fields: A section containing a few portfolio fields that will be imported when wbportfolio is installed. Can be empty.

        Returns:
            Display: The display instance
        """

        employees_section = Section(
            key="employees_section",
            collapsible=False,
            title=_("Employees"),
            display=Display(
                pages=[
                    Page(
                        title=_("Employees"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["employees_table"]],
                                inlines=[Inline(key="employees_table", endpoint="employees")],
                            )
                        },
                    ),
                ]
            ),
        )

        display = Display(
            pages=[
                Page(
                    title=_("Main Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["profile_image", "name", "customer_status", "activity_table"],
                                ["profile_image", "primary_telephone", "primary_telephone", "activity_table"],
                                ["profile_image", "type", "tier", "activity_table"],
                                ["profile_image", "activity_heat", "activity_heat", "activity_table"],
                                ["employees_section", "employees_section", "employees_section", "employees_section"],
                            ],
                            grid_template_columns=[
                                Style.MIN_CONTENT,
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                                "50%",
                            ],
                            grid_template_rows=[Style.rem(6), Style.rem(6), Style.rem(6)],
                            grid_auto_rows=Style.MIN_CONTENT,
                            sections=[employees_section],
                            inlines=[
                                Inline(
                                    key="activity_table",
                                    endpoint="wbcrm_activity",
                                    title="Activities",
                                    hide_controls=True,
                                ),
                            ],
                        ),
                    },
                ),
                Page(
                    title=_("Additional Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["salutation", "."], ["description", "description"]],
                            grid_template_columns=["minmax(min-content, 1fr)", "minmax(min-content, 2fr)"],
                            grid_auto_rows=Style.MIN_CONTENT,
                        )
                    },
                ),
            ]
        )

        # Need to insert the portfolio fields into the display
        for page in display.pages:
            if page.title == "Main Information":
                for layout_key in page.layouts.keys():
                    if portfolio_fields:
                        # Insert the section with the AUM fields below the profile picture at the left
                        page.layouts[layout_key].grid_template_areas.insert(
                            4,
                            [
                                portfolio_fields.key,
                                portfolio_fields.key,
                                portfolio_fields.key,
                                "activity_table",
                            ],
                        )
                        page.layouts[layout_key].sections.append(portfolio_fields)
                    if aum_table:
                        # Insert the section with the AUM table at the bottom right
                        page.layouts[layout_key].grid_template_areas[-1][-1] = aum_table.key
                        page.layouts[layout_key].sections.append(aum_table)
                break
        return display

    @classmethod
    def _get_new_company_instance_display(cls) -> Display:
        """Returns the display for creating a new company

        Returns:
            Display: The display instance
        """

        return Display(
            pages=[
                Page(
                    title=_("Main Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["profile_image", "name", "customer_status", "."],
                                ["profile_image", "primary_email", "primary_telephone", "."],
                                ["profile_image", "primary_manager", "type", "."],
                            ],
                            grid_template_columns=[
                                Style.MIN_CONTENT,
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                                Style.fr(1),
                            ],
                            grid_auto_rows=Style.MIN_CONTENT,
                            gap=Style.rem(5),
                        ),
                    },
                ),
                Page(
                    title=_("Additional Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["salutation", "."], ["description", "description"]],
                            grid_template_columns=["minmax(min-content, 1fr)", "minmax(min-content, 2fr)"],
                            grid_auto_rows=Style.MIN_CONTENT,
                            gap=Style.rem(5),
                        )
                    },
                ),
            ],
        )

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="cities", label=_("City")),
                dp.Field(key="type", label=_("Type")),
                dp.Field(key="tier", label=_("Tier")),
                dp.Field(key="customer_status", label=_("Status")),
                dp.Field(key="primary_manager_repr", label=_("Primary Relationship Manager")),
                dp.Field(
                    key=None,
                    label=_("Last Event"),
                    children=[
                        dp.Field(key="last_event", label=_("Name")),
                        dp.Field(key="last_event_period_endswith", label=_("Date")),
                    ],
                ),
                dp.Field(key="activity_heat", label=_("Activity Heat")),
            ]
        )

    AUM_TABLE: Section = []
    PORTFOLIO_FIELDS: Section = []

    def get_instance_display(self) -> Display:
        return (
            self._get_company_instance_display(self.AUM_TABLE, self.PORTFOLIO_FIELDS)
            if "pk" in self.view.kwargs
            else self._get_new_company_instance_display()
        )


class BankModelDisplay(CompanyModelDisplay):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="primary_manager_repr", label=_("Primary Relationship Manager")),
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="primary_telephone", label=_("Telephone")),
                dp.Field(key="primary_email", label=_("Email")),
                dp.Field(key="primary_address", label=_("Address")),
            ]
        )


class UserIsManagerDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="computed_str", label=_("Name")),
                dp.Field(key="primary_telephone", label=_("Primary Telephone")),
                dp.Field(key="primary_email", label=_("Primary Email")),
                dp.Field(key="primary_address", label=_("Primary Address")),
            ]
        )



from dataclasses import dataclass, field
from typing import Callable

from django.utils.functional import cached_property
from django.utils.module_loading import autodiscover_modules, import_string


class CachedDict(dict):
    def __init__(self):
        super().__init__()
        self._cache = {}

    def __getitem__(self, key):
        if key in self._cache:
            return self._cache[key]
        value = super().get(key)
        self._cache[key] = import_string(value)
        return self._cache[key]


@dataclass
class WorkflowSite:
    registered_assignees_methods: dict[str, Callable] = field(default_factory=dict)
    registered_assignees_names: dict[str, str] = field(default_factory=dict)
    registered_model_classes_serializer_map: CachedDict = field(default_factory=CachedDict)

    @cached_property
    def assignees_choices(self) -> tuple[str, str]:
        autodiscover_modules("workflows")  # we have to do that because the filter choices loads at runtime
        return tuple(self.registered_assignees_names.items())

    @cached_property
    def assignees_methods(self) -> tuple[str, Callable]:
        return tuple(self.registered_assignees_methods.items())

    @cached_property
    def model_content_types(self):
        return [
            (model._meta.app_label, model.__name__) for model in self.registered_model_classes_serializer_map.keys()
        ]


workflow_site = WorkflowSite()



from collections import OrderedDict

from django.utils.translation import gettext as _
from rest_framework.reverse import NoReverseMatch, reverse

from wbcore import serializers
from wbcore.contrib.agenda.models import CalendarItem


class CalendarItemRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.SerializerMethodField()

    def get__detail(self, obj) -> str:
        try:
            return reverse(f"{obj.endpoint_basename}-detail", args=[obj.id], request=self.context["request"])
        except (NoReverseMatch, KeyError):
            return ""

    class Meta:
        model = CalendarItem
        fields = (
            "id",
            "title",
            "_detail",
        )


class CalendarItemModelSerializer(serializers.ModelSerializer):
    activity_id = serializers.CharField()
    endpoint = serializers.SerializerMethodField()
    is_confidential = serializers.BooleanField(default=False, read_only=True)
    is_draggable = serializers.BooleanField(default=False, read_only=True)
    is_private = serializers.BooleanField(default=False, read_only=True)

    def get_endpoint(self, obj):
        try:
            return reverse(f"{obj.endpoint_basename}-detail", args=[obj.id], request=self.context["request"])
        except (NoReverseMatch, KeyError):
            return ""

    def to_representation(self, instance: CalendarItem) -> OrderedDict:
        representation = super().to_representation(instance)

        is_private: bool
        if (is_private := representation.get("is_private", False)) or representation.get("is_confidential", False):
            item_type = instance.item_type.split(".")[1] if instance.item_type else _("Item")
            representation["title"] = (
                _("Private {}").format(item_type) if is_private else _("Confidential {}").format(item_type)
            )
            if not is_private:
                representation["_entities"] = ""
                representation["entities"] = ""
                representation["entity_list"] = ""

        return representation

    class Meta:
        model = CalendarItem
        read_only_fields = ("is_cancelled",)
        fields = (
            "id",
            "title",
            "entities",
            "all_day",
            "activity_id",
            "color",
            "endpoint",
            "entity_list",
            "icon",
            "is_cancelled",
            "is_confidential",
            "is_draggable",
            "is_private",
            "period",
            "is_deletable",
        )
        read_only_fields = ("endpoint",)



from .user_activities import (
    UserActivityModelDisplay,
    UserActivityModelUserDisplay,
    UserActivityTableDisplayConfig,
)
from .users import UserModelDisplay, UserPermissionModelDisplay, UserProfileModelDisplay



from django.contrib.contenttypes.models import ContentType
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response

from wbcore import viewsets
from wbcore.contrib.i18n.translation import translate_model_as_task


class ModelTranslateMixin(viewsets.ModelViewSet):
    @action(methods=["POST"], detail=True)
    def auto_translate(self, request: Request, *args, **kwargs):
        """Initiates an asynchronous translation task for the specified model instance.

        This method retrieves the model instance based on the provided URL parameters,
        starts a background translation task, and returns a notification response.

        Args:
            request (Request): The HTTP request object containing the data for the translation.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
            override_existing_data (bool): If True, existing content will not be overridden.

        Returns:
            Response: A response indicating the status of the translation initiation.
        """
        override_existing_data = request.data.get("override_existing_data", "false") == "true"

        obj = self.get_object()
        if obj:
            ct = ContentType.objects.get_for_model(obj)  # type: ignore
            translate_model_as_task.delay(ct.id, obj.id, override_existing_data)
            return Response({"__notification": "The translation started in the background."})

        return Response({"non_field_errors": ["The URL was malformatted."]}, status=status.HTTP_400_BAD_REQUEST)



# Generated by Django 5.0.10 on 2025-01-28 15:57

from django.db import migrations

def migrate_city(apps, schema_editor):
    AddressContact = apps.get_model('directory', 'AddressContact')
    from wbcore.contrib.geography.models import  Geography
    qs = AddressContact.objects.filter(geography_city__isnull=True, city__isnull=False)
    from tqdm import tqdm
    for a in tqdm(qs, total=qs.count()):
        if city := Geography.cities.get_by_name(a.city):
            a.geography_city_id = city.id
        else:
            a.province = f"{a.province} {a.city}"  if a.province else a.city # we add to an temporary field to not lose this information
        a.save()

class Migration(migrations.Migration):

    dependencies = [
        ('directory', '0009_remove_entry_external_identfier_and_more'),
    ]

    operations = [
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(migrate_city),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.RemoveField(
            model_name='addresscontact',
            name='city',
        ),
    ]



from rest_framework import filters

from wbcore import viewsets
from wbcore.filters import DjangoFilterBackend

from ..models import CurrencyFXRates
from ..serializers import CurrencyFXRatesModelSerializer
from .display import CurrencyFXRatesCurrencyDisplayConfig
from .endpoints import CurrencyFXRatesCurrencyEndpointConfig
from .titles import CurrencyFXRatesCurrencyTitleConfig


class CurrencyFXRatesCurrencyModelViewSet(viewsets.ModelViewSet):
    display_config_class = CurrencyFXRatesCurrencyDisplayConfig
    endpoint_config_class = CurrencyFXRatesCurrencyEndpointConfig
    title_config_class = CurrencyFXRatesCurrencyTitleConfig

    serializer_class = CurrencyFXRatesModelSerializer
    filter_backends = (filters.OrderingFilter, DjangoFilterBackend)

    ordering = ordering_fields = ["-date"]

    filter_backends = (DjangoFilterBackend, filters.OrderingFilter)
    queryset = CurrencyFXRates.objects.all()

    filterset_fields = {"date": ["gte", "exact", "lte"], "value": ["gte", "exact", "lte"]}
    ordering_fields = ("date", "value")

    ordering = ["-date"]

    def get_queryset(self):
        _id = self.kwargs["currency_id"]
        return CurrencyFXRates.objects.filter(currency__id=_id).all()



from django.dispatch import receiver
from rest_framework.reverse import reverse

from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.signals.instance_buttons import add_extra_button


@receiver(add_extra_button)
def add_extra_documents(sender, instance, request, view, pk=None, **kwargs):
    if instance and pk and view:
        content_type = view.get_content_type()
        endpoint = reverse("wbcore:documents:document_content_object", args=[content_type.id, pk], request=request)
        return bt.WidgetButton(endpoint=endpoint, label="Documents", icon=WBIcon.DOCUMENT_WITH_ATTACHMENT.icon)



import warnings

warnings.warn(
    "The 'wbcore.pandas' module is deprecated and will be removed in a future version. "
    "Please use 'wbcore.contrib.pandas' instead.",
    DeprecationWarning,
    stacklevel=2,
)

from wbcore.contrib.pandas import fields
from wbcore.contrib.pandas.fields import (  # noqa
    PKField,
    CharField,
    DateField,
    DateRangeField,
    BooleanField,
    TextField,
    EmojiRatingField,
    FloatField,
    IntegerField,
    YearField,
    ListField,
    JsonField,
    SparklineField,
    PandasFields,
)
from wbcore.contrib.pandas.filters import (  # noqa
    PandasDjangoFilterBackend,
    PandasSearchFilter,
    PandasOrderingFilter,
)
from wbcore.contrib.pandas.filterset import PandasFilterSetMixin  # noqa
from wbcore.contrib.pandas.metadata import PandasMetadata  # noqa
from wbcore.contrib.pandas.utils import (  # noqa
    rule,
    overwrite,
    overwrite_rule,
    overwrite_row,
    overwrite_row_df,
    override_number_to_percent,
    override_number_with_currency,
    override_number_to_x,
    override_number_precision,
    override_number_to_decimal,
    override_number_to_integer_without_decorations,
    pct_change_with_negative_values,
    get_empty_series,
    sanitize_field,
    sanitize_fields,
)
from wbcore.contrib.pandas import views
from wbcore.contrib.pandas.views import PandasMixin, PandasAPIView, PandasAPIViewSet  # noqa



from rest_framework.metadata import SimpleMetadata


class ButtonMetadata(SimpleMetadata):
    def generate_metadata(self, request, view):
        if hasattr(super(), "generate_metadata"):
            yield from super().generate_metadata(request, view)
        yield "buttons", [btn.serialize(request=request) for btn in view.get_buttons(request)]



from .condition import ConditionFactory
from .data import DataFactory, DataValueFactory
from .display import DisplayFactory
from .process import ProcessFactory, ProcessStepFactory
from .step import (
    DecisionStepFactory,
    EmailStepFactory,
    FinishStepFactory,
    JoinStepFactory,
    RandomChildStepFactory,
    ScriptStepFactory,
    SplitStepFactory,
    StepFactory,
    UserStepFactory,
    StartStepFactory,
)
from .transition import TransitionFactory
from .workflow import WorkflowFactory



import sys
import uuid
from contextlib import suppress
from pathlib import Path

from django.apps import apps
from django.conf import settings
from django.contrib.postgres.aggregates import ArrayAgg
from django.core.exceptions import AppRegistryNotReady
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import CharField, Exists, F, OuterRef, Q, Subquery, Value
from django.db.models.functions import Concat
from django.db.models.signals import m2m_changed
from django.dispatch import receiver
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from dynamic_preferences.registries import global_preferences_registry
from modeltrans.fields import TranslationField
from slugify import slugify

from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.authentication.models import User
from wbcore.contrib.directory.models.contacts import (
    AddressContact,
    BankingContact,
    ContactLocationChoices,
    EmailContact,
    TelephoneContact,
    WebsiteContact,
)
from wbcore.contrib.directory.models.relationships import ClientManagerRelationship, EmployerEmployeeRelationship
from wbcore.contrib.directory.signals import deactivate_profile
from wbcore.contrib.directory.typings import Person as PersonDTO
from wbcore.models import WBModel
from wbcore.permissions.shortcuts import get_internal_users
from wbcore.utils.models import (
    ActiveObjectManager,
    ComplexToStringMixin,
    DeleteToDisableMixin,
)


class CustomerStatus(WBModel):
    title = models.CharField(
        max_length=32,
        verbose_name=_("Title"),
        unique=True,
        blank=False,
        null=False,
    )

    slugify_title = models.CharField(
        max_length=32,
        unique=True,
        verbose_name="Slugified Title",
        blank=True,
        null=True,
    )

    def __str__(self) -> str:
        return f"{self.title}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:customerstatus"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:customerstatusrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"

    class Meta:
        verbose_name = _("Customer Status")
        verbose_name_plural = _("Customer Statuses")

    def save(self, *args, **kwargs):
        self.slugify_title = slugify(self.title, separator=" ")
        super().save(*args, **kwargs)


class CompanyType(WBModel):
    title = models.CharField(
        max_length=128,
        verbose_name=_("Title"),
        unique=True,
        blank=False,
        null=False,
    )

    slugify_title = models.CharField(
        max_length=128,
        unique=True,
        verbose_name="Slugified Title",
        blank=True,
        null=True,
    )

    def __str__(self) -> str:
        return f"{self.title}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:companytype"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:companytyperepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"

    class Meta:
        verbose_name = _("Company Type")
        verbose_name_plural = _("Company Types")

    def save(self, *args, **kwargs):
        self.slugify_title = slugify(self.title, separator=" ")
        super().save(*args, **kwargs)


class Specialization(WBModel):
    title = models.CharField(
        max_length=128,
        verbose_name=_("Title"),
        unique=True,
        blank=False,
        null=False,
    )

    slugify_title = models.CharField(
        max_length=128,
        unique=True,
        verbose_name="Slugified Title",
        blank=True,
        null=True,
    )

    def __str__(self) -> str:
        return f"{self.title}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:specialization"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:specializationrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"

    class Meta:
        verbose_name = _("Specialization")
        verbose_name_plural = _("Specializations")

    def save(self, *args, **kwargs):
        self.slugify_title = slugify(self.title, separator=" ")
        super().save(*args, **kwargs)


def upload_to_profile_images(instance, filename):
    file_extension = "".join(Path(filename).suffixes)
    return f"crm/entry/profile_images/{instance.uuid}{file_extension}"


def upload_to_signatures(instance, filename):
    file_extension = "".join(Path(filename).suffixes)
    return f"crm/entry/signatures/{instance.uuid}{file_extension}"


class EntryDefaultQueryset(models.QuerySet):
    def filter_for_user(self, user: User) -> models.QuerySet:
        """
        Protect the chained queryset and filter the entries that this user cannot see based on the following rules:

        * internal user or supersuer: All
        * external user: only itself, its colleagues and all its clients
        * else: None
        """
        profile = user.profile
        if user.is_superuser or profile.is_internal:
            return self
        else:
            return (
                self.annotate(
                    has_valid_relationship_manager_role=Exists(
                        ClientManagerRelationship.objects.filter(
                            client__id=OuterRef("pk"), relationship_manager=profile
                        ).exclude(status=ClientManagerRelationship.Status.REMOVED)
                    ),
                    is_colleague=Exists(
                        EmployerEmployeeRelationship.objects.filter(
                            employee_id=OuterRef("pk"), employer__in=profile.employers.all()
                        )
                    ),
                )
                .filter(
                    Q(is_colleague=True)  # can see its colleagues
                    | Q(id__in=profile.employers.values("id"))  # can see its colleagues
                    | Q(id=profile.id)  # can see itself
                    | Q(
                        has_valid_relationship_manager_role=True
                    )  # can see all entry where they have a valid relationship manager role
                )
                .distinct()
            )

    def filter_only_internal(self) -> models.QuerySet:
        return self.filter(id__in=get_internal_users().values("profile"))

    def annotate_all(self) -> models.QuerySet:
        qs = self
        with suppress(AppRegistryNotReady):
            qs = qs.annotate(
                primary_manager_repr=Subquery(
                    ClientManagerRelationship.objects.filter(
                        client__id=OuterRef("id"),
                        primary=True,
                        status__in=[
                            ClientManagerRelationship.Status.APPROVED,
                            ClientManagerRelationship.Status.PENDINGREMOVE,
                        ],
                    ).values("relationship_manager__computed_str")[:1]
                ),
                primary_email=Subquery(
                    EmailContact.objects.filter(primary=True, entry__id=OuterRef("pk")).values("address")[:1],
                    output_field=CharField(),
                ),
                secondary_email=Subquery(
                    EmailContact.objects.filter(primary=False, entry__id=OuterRef("pk")).values("address")[:1],
                    output_field=CharField(),
                ),
                primary_telephone=Subquery(
                    TelephoneContact.objects.filter(primary=True, entry__id=OuterRef("pk")).values("number")[:1],
                    output_field=CharField(),
                ),
                primary_address=Subquery(
                    AddressContact.objects.filter(primary=True, entry__id=OuterRef("pk"))
                    .annotate(
                        primary_address=Concat(
                            F("street"), Value(" "), F("zip"), Value(" "), F("geography_city__name")
                        )
                    )
                    .values("primary_address")[:1],
                    output_field=CharField(),
                ),
                last_event=Subquery(
                    CalendarItem.objects.filter(
                        period__endswith__lte=timezone.now(),
                        visibility=CalendarItem.Visibility.PUBLIC,
                        entities=OuterRef("pk"),
                    )
                    .order_by("-period__startswith")
                    .values("title")[:1]
                ),
                last_event_period_endswith=Subquery(
                    CalendarItem.objects.filter(
                        period__endswith__lte=timezone.now(),
                        visibility=CalendarItem.Visibility.PUBLIC,
                        entities=OuterRef("pk"),
                    )
                    .order_by("-period__startswith")
                    .values("period__endswith")[:1]
                ),
                cities=ArrayAgg("addresses__geography_city", filter=Q(addresses__geography_city__isnull=False)),
            )
        return qs


class EntryManager(ActiveObjectManager):
    def get_queryset(self) -> EntryDefaultQueryset:
        return EntryDefaultQueryset(self.model).filter(is_active=True)

    def filter_for_user(self, user: User) -> models.QuerySet:
        return self.get_queryset().filter_for_user(user)

    def filter_only_internal(self) -> models.QuerySet:
        return self.get_queryset().filter_only_internal()

    def annotate_all(self):
        return self.get_queryset().annotate_all()


class Entry(ComplexToStringMixin, DeleteToDisableMixin, WBModel):
    AUTOMATICALLY_CLEAN_SOFT_DELETED_OBJECTS = True

    class EntryType(models.TextChoices):
        PERSON = "PERSON", _("Person")
        COMPANY = "COMPANY", _("Company")

    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)
    entry_type = models.CharField(max_length=255, default="", choices=EntryType.choices, verbose_name=_("Type"))
    additional_fields = models.JSONField(default=dict, blank=True, null=True, verbose_name=_("Additional Fields"))
    is_draft_entry = models.BooleanField(
        default=False,
        verbose_name=_("Draft Entry"),
        help_text=_(
            "Draft entries are entries that have been created automatically and still need verification by a human user."
        ),
    )
    relationship_managers = models.ManyToManyField(
        "Person",
        related_name="clients",
        blank=True,
        through="directory.ClientManagerRelationship",
        through_fields=("client", "relationship_manager"),
        verbose_name=_("Relationship Managers"),
        help_text=_("People in charge of this entry"),
    )
    relationships = models.ManyToManyField(
        blank=True,
        symmetrical=False,
        to="self",
        through="directory.Relationship",
        through_fields=("from_entry", "to_entry"),
        verbose_name=_("The Entry's Relationships"),
    )
    # Activity heat describes how active a customer is. The more active a customer, the higher their activity heat.
    activity_heat = models.FloatField(
        validators=[MinValueValidator(0), MaxValueValidator(1)],
        default=0,
        verbose_name=_("Activity Heat"),
    )
    slugify_computed_str = models.CharField(
        max_length=512, blank=True, null=True, verbose_name="Slugified Search Name"
    )
    profile_image = models.ImageField(blank=True, null=True, upload_to=upload_to_profile_images, max_length=256)
    # TODO: Move to Person Model?
    signature = models.ImageField(blank=True, null=True, upload_to=upload_to_signatures, max_length=256)
    external_identifier = models.CharField(null=True, blank=True, max_length=128, unique=True)
    salutation = models.CharField(max_length=255, verbose_name=_("Salutation"), null=True, blank=True)

    telephones: models.QuerySet[TelephoneContact]
    emails: models.QuerySet[EmailContact]
    addresses: models.QuerySet[AddressContact]
    websites: models.QuerySet[WebsiteContact]
    banking: models.QuerySet[BankingContact]

    class Meta:
        verbose_name = _("Entry")
        verbose_name_plural = _("Entries")

        ordering = ("entry_type",)

    objects = EntryManager()

    def __str__(self):
        return self.computed_str

    @property
    def is_company(self) -> bool:
        return self.entry_type == Company.__name__

    @staticmethod
    def _primary_contact(ref: models.QuerySet) -> models.Model | None:
        with suppress(ref.model.DoesNotExist):
            return ref.get(primary=True)

    def primary_email_contact(self):
        return self._primary_contact(self.emails)

    def primary_telephone_contact(self):
        return self._primary_contact(self.telephones)

    def primary_address_contact(self):
        return self._primary_contact(self.addresses)

    def primary_website_contact(self):
        return self._primary_contact(self.websites)

    def primary_banking_contact(self):
        return self._primary_contact(self.banking)

    def get_casted_entry(self):
        """
        Cast the entry into its child representative. We use the all manager to be sure to get the casted object even after its "deletion"
        """
        try:
            model_class = getattr(sys.modules[__name__], self.entry_type)
        except AttributeError:
            model_class = self.__class__
        return model_class.all_objects.get(pk=self.pk)

    def compute_str(self):
        self.slugify_computed_str = slugify(_("Entry {id}").format(id=self.id), separator=" ")
        return _("Entry {id}").format(id=self.id)

    def delete_additional_fields(self, additional_field_key=None):
        if additional_field_key in self.additional_fields:
            del self.additional_fields[additional_field_key]

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:entry"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:entryrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"


class PersonDefaultQueryset(EntryDefaultQueryset):
    def annotate_all(self) -> models.QuerySet:
        qs = super().annotate_all()
        with suppress(AppRegistryNotReady):
            qs = qs.annotate(
                primary_employer=Subquery(
                    EmployerEmployeeRelationship.objects.filter(employee__id=OuterRef("pk"), primary=True).values(
                        "employer__id"
                    )[:1]
                ),
                primary_employer_repr=Subquery(
                    EmployerEmployeeRelationship.objects.filter(primary=True, employee=OuterRef("pk")).values(
                        "employer__computed_str"
                    )[:1]
                ),
                primary_employer_number=Subquery(
                    TelephoneContact.objects.filter(primary=True, entry__id=OuterRef("primary_employer")).values(
                        "number"
                    )[:1],
                    output_field=CharField(),
                ),
                name=Concat(F("first_name"), Value(" "), F("last_name")),
                customer_status=Subquery(
                    EmployerEmployeeRelationship.objects.filter(primary=True, employee=OuterRef("pk")).values(
                        "employer__customer_status__title"
                    )[:1]
                ),
                position_in_company=Subquery(
                    EmployerEmployeeRelationship.objects.filter(primary=True, employee=OuterRef("pk")).values(
                        "position__title"
                    )[:1]
                ),
                tier=Subquery(
                    EmployerEmployeeRelationship.objects.filter(
                        primary=True,
                        employee=OuterRef("pk"),
                    ).values("employer__tier")[:1]
                ),
            )
        return qs


class DefaultPersonManager(EntryManager):
    def get_queryset(self):
        qs = PersonDefaultQueryset(self.model).filter(is_active=True)
        with suppress(Exception):
            qs = qs.exclude(user_account__email=getattr(settings, "ANONYMOUS_USER_NAME", "AnonymousUser"))
        return qs


class RegisteredPersonManager(DefaultPersonManager):
    """Custom Manager for filtering directly for registered user or non user"""

    def get_queryset(self):
        qs = super().get_queryset()
        with suppress(Exception):
            qs = qs.exclude(Q(user_account__isnull=False) & Q(user_account__is_register=False))
        return qs


class PersonEmployeeManager(RegisteredPersonManager):
    """Custom Manager for filtering directly for Employees"""

    def get_queryset(self):
        if apps.is_installed("wbhuman_resources"):
            return (
                super()
                .get_queryset()
                .filter(
                    Q(human_resources__isnull=False)
                    & Q(human_resources__is_active=True)
                    & Q(user_account__isnull=False)
                )
            )
        else:
            try:
                main_company_id = global_preferences_registry.manager()["directory__main_company"]
                return super().get_queryset().filter(employers__id=main_company_id).distinct()
            except Exception:
                return super().get_queryset().none()


class Person(Entry):
    class Prefix(models.TextChoices):
        MR = "MR", _("Mr.")
        MRS = "MRS", _("Mrs.")
        PROF = "PROF", _("Prof.")
        DR = "DR", _("Dr.")
        MED = "MED", _("Med.")
        ME = "ME", _("Me.")

    """Person Model that represents a person in the CRM"""
    prefix = models.CharField(max_length=4, choices=Prefix.choices, verbose_name=_("Prefix"))

    first_name = models.CharField(max_length=255, verbose_name=_("First Name"))
    last_name = models.CharField(max_length=255, verbose_name=_("Last Name"))
    initials = models.CharField(max_length=4, blank=True, null=True, verbose_name=_("Initials"))

    employers = models.ManyToManyField(
        "Company",
        related_name="employees",
        blank=True,
        through="directory.EmployerEmployeeRelationship",
        through_fields=("employee", "employer"),
        verbose_name=_("Employers"),
        help_text=_("The person's employers"),
    )

    active_employee = models.BooleanField(default=True, verbose_name=_("Available Employee"))
    birthday = models.DateField(null=True, blank=True, verbose_name=_("Birthday"))
    personality_profile_red = models.FloatField(
        validators=[MinValueValidator(0), MaxValueValidator(1)],
        default=0,
    )
    personality_profile_green = models.FloatField(
        validators=[MinValueValidator(0), MaxValueValidator(1)],
        default=0,
    )
    personality_profile_blue = models.FloatField(
        validators=[MinValueValidator(0), MaxValueValidator(1)],
        default=0,
    )

    formal = models.BooleanField(default=True, verbose_name=_("Formal"))
    specializations = models.ManyToManyField(
        to="Specialization",
        blank=True,
        verbose_name=_("Specializations"),
    )
    description = models.TextField(default="", blank=True)
    i18n = TranslationField(fields=["description"])

    objects = DefaultPersonManager()
    registered_users = RegisteredPersonManager()

    class Meta:
        verbose_name = _("Person")
        verbose_name_plural = _("Persons")

    @cached_property
    def is_internal(self) -> bool:
        if user := getattr(self, "user_account", None):
            return user.is_internal
        return False

    @cached_property
    def full_name(self):
        return f"{self.last_name} {self.first_name}"

    def __str__(self):
        return f"{self.first_name} {self.last_name}"

    def _build_dto(self):
        return PersonDTO(
            first_name=self.first_name,
            last_name=self.last_name,
            email=contact.address if (contact := self.primary_email_contact()) else None,
            id=self.id,
        )

    def str_full(self):
        """
        Get the string representation including Employers seperated title
        """
        employers_repr = None
        eer_rel = EmployerEmployeeRelationship.objects.filter(employee=self.id, primary=True)
        try:
            if eer_rel.exists():
                employers_repr = Company.all_objects.get(id=eer_rel[0].employer.id).name
            else:
                employers_repr = (
                    "/".join(self.employers.all().values_list("name", flat=True)) if self.employers.exists() else ""
                )
            if employers_repr:
                return f"{self.first_name} {self.last_name}{(' (%s)' % employers_repr)}"
        except ValueError:
            pass
        return f"{self.first_name} {self.last_name}"

    def compute_str(self):
        self.slugify_computed_str = slugify(self.str_full(), separator=" ")
        return self.str_full()

    def get_initials(self):
        initials = []
        if self.first_name and len(self.first_name) > 1:
            n = 1 if self.last_name else 2
            initials.append(self.first_name.upper()[:n])

        if self.last_name and len(self.last_name) > 1:
            n = 1 if self.first_name else 2
            initials.append(self.last_name.upper()[:n])

        if len(initials) == 0:
            initials.append("n.a.")

        return "".join(initials)

    def save(self, *args, **kwargs):
        self.entry_type = "Person"
        self.initials = self.get_initials()
        if not self.salutation:
            self.salutation = global_preferences_registry.manager()["directory__person_salutation"].format(
                self.first_name, self.last_name
            )
        super().save(*args, **kwargs)

    @classmethod
    def get_or_create_with_user(cls, user, first_name="", last_name=""):
        """
        Helper method that creates the CRM profile from user

        Arguments:
            user {User} -- The user used to generate the profile
        """

        if profile := getattr(user, "profile", None):
            return profile
        else:
            associated_emails = EmailContact.objects.filter(
                address=user.email, entry__isnull=False, entry__entry_type="Person", entry__is_active=True
            )
            if (
                associated_emails.exists()
                and (person := Person.objects.get(id=associated_emails.first().entry.id))
                and getattr(person, "user_account", None) is None
            ):
                return person
            else:
                usernames = user.username.split("-")
                if not first_name:
                    first_name = usernames[0]
                if not last_name and len(usernames) > 1:
                    last_name = usernames[1]
                person = cls.objects.create(first_name=first_name, last_name=last_name)
                EmailContact.objects.create(
                    primary=True,
                    location=ContactLocationChoices.WORK.name,
                    entry=person,
                    address=user.email,
                )
            return person

    @classmethod
    def create_with_attributes(cls, first_name, last_name, email=None):
        """
        Helper method that creates the CRM profile from user

        Arguments:
            user {User} -- The user used to generate the profile
        """

        person = cls.objects.create(first_name=first_name, last_name=last_name)
        if email:
            EmailContact.objects.create(
                primary=True,
                location=ContactLocationChoices.WORK.name,
                entry=person,
                address=email,
            )

        return person

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:person"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:personrepresentation-list"

    @classmethod
    def get_relationship_managers_representation_endpoint(cls):
        return "wbcore:directory:personinchargerepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"


class BankCompanyManager(EntryManager):
    """Custom Manager for filtering directly for Investment Banks"""

    def get_queryset(self):
        if hasattr(self, "issues_products"):
            return super().get_queryset().filter(issues_products__isnull=False).distinct()

        # This model manager is called from other places on startup time,
        # hence we suppress the programming error if the migration has not yet run
        return super().get_queryset().filter(type__title="Bank").distinct()


class Company(Entry):
    class Tiering(models.TextChoices):
        ONE = "ONE", "1"
        TWO = "TWO", "2"
        THREE = "THREE", "3"
        FOUR = "FOUR", "4"
        FIVE = "FIVE", "5"

    """
    Company Model that represents a Company in the CRM
    """

    type = models.ForeignKey(
        "CompanyType",
        related_name="company",
        on_delete=models.SET_NULL,
        verbose_name=_("Type"),
        null=True,
        blank=False,
    )

    name = models.CharField(max_length=255, verbose_name=_("Name"))

    objects = EntryManager()

    customer_status = models.ForeignKey(
        "CustomerStatus",
        related_name="company",
        on_delete=models.SET_NULL,
        verbose_name=_("Customer Status"),
        null=True,
        blank=True,
    )

    headcount = models.CharField(max_length=32, default="", blank=True, verbose_name=_("Number of employees"))
    tier = models.CharField(max_length=16, null=True, blank=True, choices=Tiering.choices, verbose_name=_("Tier"))

    description = models.TextField(default="", blank=True, verbose_name=_("Description"))

    class Meta:
        verbose_name = _("Company")
        verbose_name_plural = _("Companies")

    def __str__(self):
        return f"{self.name}"

    def save(self, *args, **kwargs):
        self.entry_type = "Company"
        if not self.salutation:
            self.salutation = global_preferences_registry.manager()["directory__company_salutation"].format(self.name)
        super().save(*args, **kwargs)

    def compute_str(self):
        self.slugify_computed_str = slugify(self.name, separator=" ")
        return self.name

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:company"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:companyrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"


@receiver(models.signals.post_save, sender=Entry)
def post_save_base_info(sender, instance, created, raw, **kwargs):
    if not raw:
        casted_entry = instance.get_casted_entry()
        Entry.objects.filter(id=instance.id).update(
            entry_type=casted_entry.__class__.__name__, computed_str=casted_entry.compute_str()
        )


@receiver(models.signals.post_save, sender=Company)
def post_save_company(sender, instance, created, raw, **kwargs):
    """
    Company post save signal: Triggers the post_save signals of all its employees, which updates their computed_str.
    """
    if not raw:
        if instance.employees.exists():
            for employee in instance.employees.all():
                employee.save()


@receiver(m2m_changed, sender=Person.employers.through)
def notify_entry_for_potential_claim(sender, instance, action, **kwargs):
    """
    m2m change signal: If an employer is removed or added from the person employeers M2M, we need to recompute the computed_str
    """
    if not instance.is_company:
        instance.employers.all()
        if action in ["post_add", "post_remove", "post_clear"]:
            new_computed_str = instance.str_full()
            Person.objects.filter(id=instance.id).update(computed_str=new_computed_str)


@receiver(deactivate_profile)
def handle_user_deactivation(sender, instance, substitute_profile=None, **kwargs):
    messages = []
    main_company_id = global_preferences_registry.manager()["directory__main_company"]
    try:
        main_company = Company.objects.get(id=main_company_id)

        if main_company in instance.employers.all():
            instance.employers.remove(main_company)
            messages.append(
                _("Removed {main_company} from {profile}'s employers").format(
                    main_company=main_company.computed_str, profile=instance.computed_str
                )
            )
    except Company.DoesNotExist:
        pass
    users_clients = ClientManagerRelationship.objects.filter(relationship_manager=instance).exclude(
        status=ClientManagerRelationship.Status.REMOVED
    )
    if users_clients.exists():
        if substitute_profile:
            client_number = users_clients.count()
            for relationship in users_clients:
                substitute_relationships = ClientManagerRelationship.objects.filter(
                    relationship_manager=substitute_profile, client=relationship.client
                )

                if relationship.status in [
                    ClientManagerRelationship.Status.DRAFT,
                    ClientManagerRelationship.Status.PENDINGADD,
                    ClientManagerRelationship.Status.PENDINGREMOVE,
                ]:
                    if substitute_relationships.exists():
                        relationship.delete()
                    else:
                        relationship.relationship_manager = substitute_profile
                        relationship.save()
                elif relationship.status == ClientManagerRelationship.Status.APPROVED:
                    if substitute_relationships.exists():
                        for rel in substitute_relationships.all():
                            rel.primary = relationship.primary
                            rel.status = ClientManagerRelationship.Status.APPROVED
                            rel.save()
                    else:
                        ClientManagerRelationship.objects.create(
                            relationship_manager=substitute_profile,
                            client=relationship.client,
                            status=ClientManagerRelationship.Status.APPROVED,
                            primary=relationship.primary,
                        )
                    relationship.status = ClientManagerRelationship.Status.REMOVED
                    relationship.primary = False
                    relationship.save()

            messages.append(
                _("Assigned {clients} manager role(s) to {substitute_profile}").format(
                    clients=client_number, substitute_profile=substitute_profile.computed_str
                )
            )

        else:
            for rel in users_clients.all():
                rel.status = ClientManagerRelationship.Status.REMOVED
                rel.primary = False
                rel.save()
    return messages



import pytest
from django.test import Client, TestCase
from django.urls import reverse
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.example_app.factories import EventFactory, EventTypeFactory
from wbcore.contrib.example_app.models import Event, EventType
from wbcore.contrib.example_app.serializers import (
    EventModelSerializer,
    EventTypeModelSerializer,
)
from wbcore.contrib.example_app.tests.test_viewsets.test_utils_viewsets import (
    find_instances_in_response,
    get_create_view,
    get_delete_view,
    get_detail_view,
    get_partial_view,
    get_update_view,
)
from wbcore.contrib.example_app.viewsets import EventModelViewSet, EventTypeModelViewSet


@pytest.mark.django_db
class TestEventModelViewSet(TestCase):
    def setUp(self) -> None:
        self.user = UserFactory(is_active=True, is_superuser=True)
        self.client = Client()
        self.client.force_login(user=self.user)
        self.list_url = reverse("example_app:event-list")
        self.detail_url_str = "example_app:event-detail"

    def test_list_view(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

    def test_create_view(self):
        event = EventFactory()
        response = get_create_view(self.client, event, self.user, self.list_url, EventModelViewSet)
        self.assertEqual(response.status_code, 201)
        self.assertTrue(
            Event.objects.filter(event_type=event.event_type, match=event.match, minute=event.minute).exists()
        )

    def test_detail_view(self):
        instance = EventFactory()
        response = get_detail_view(self.client, instance.pk, self.detail_url_str)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["id"], instance.id)

    def test_update_view(self):
        instance = EventFactory()
        max_match_duration = instance.match.sport.match_duration
        instance.minute = max_match_duration - instance.minute
        response = get_update_view(self.client, instance, EventModelSerializer, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["minute"], instance.minute)

    def test_partial_update_view(self):
        instance = EventFactory()
        max_match_duration = instance.match.sport.match_duration
        response = get_partial_view(
            self.client, instance.id, {"minute": max_match_duration - instance.minute}, self.detail_url_str
        )
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["minute"], instance.minute)

    def test_delete_view(self):
        instance = EventFactory()
        response = get_delete_view(self.client, self.detail_url_str, instance.pk)
        # It is not possible to delete an event, since the get_endpoint_url returns None. I suppose that is expected behavior.
        self.assertEqual(response.status_code, 405)
        self.assertTrue(Event.objects.filter(pk=instance.pk).exists())

    def test_ordering_fields(self):
        event1 = EventFactory(minute=30)
        EventFactory(person=event1.person, minute=20, event_type=event1.event_type, match=event1.match)
        EventFactory(person=event1.person, minute=40, event_type=event1.event_type, match=event1.match)

        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], Event.objects.count())
        self.assertEqual(response.data["results"][0]["minute"], 20)
        self.assertEqual(response.data["results"][1]["minute"], 30)
        self.assertEqual(response.data["results"][2]["minute"], 40)


@pytest.mark.django_db
class TestEventTypeModelViewSet(TestCase):
    def setUp(self) -> None:
        self.user = UserFactory(is_active=True, is_superuser=True)
        self.client = Client()
        self.client.force_login(user=self.user)
        self.list_url = reverse("example_app:eventtype-list")
        self.detail_url_str = "example_app:eventtype-detail"

    def test_list_view(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

    def test_create_view(self):
        event_type = EventTypeFactory()
        response = get_create_view(self.client, event_type, self.user, self.list_url, EventTypeModelViewSet)
        self.assertEqual(response.status_code, 201)
        self.assertTrue(EventType.objects.filter(name=event_type.name).exists())

    def test_detail_view(self):
        instance = EventTypeFactory()
        response = get_detail_view(self.client, instance.pk, self.detail_url_str)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["id"], instance.id)

    def test_update_view(self):
        instance = EventTypeFactory()
        instance.name = "Updated Instance"
        response = get_update_view(self.client, instance, EventTypeModelSerializer, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_partial_update_view(self):
        instance = EventTypeFactory()
        response = get_partial_view(self.client, instance.id, {"name": "Updated Instance"}, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_delete_view(self):
        instance = EventTypeFactory()
        response = get_delete_view(self.client, self.detail_url_str, instance.pk)
        self.assertEqual(response.status_code, 204)
        self.assertFalse(Event.objects.filter(pk=instance.pk).exists())

    def test_ordering_fields(self):
        event_a = EventTypeFactory(name="BBB")
        event_b = EventTypeFactory(name="AAA", sport=event_a.sport)
        event_c = EventTypeFactory(name="CCC", sport=event_a.sport)

        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], EventType.objects.count())
        self.assertEqual(response.data["results"][0]["id"], event_b.id)
        self.assertEqual(response.data["results"][1]["id"], event_a.id)
        self.assertEqual(response.data["results"][2]["id"], event_c.id)

    def test_event_type_sport(self):
        type_a = EventTypeFactory()
        type_b = EventTypeFactory(sport=type_a.sport)
        type_c = EventTypeFactory()
        expected_number_of_types = EventType.objects.filter(sport=type_a.sport).count()
        event_type_url = reverse("example_app:eventtype-sport-list", args=[type_a.sport.id])
        response = self.client.get(event_type_url)
        type_a_found, type_b_found, type_c_found = find_instances_in_response([type_a, type_b, type_c], response)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.data["count"],
            expected_number_of_types,
            f"The answer should contain {expected_number_of_types} types",
        )
        self.assertTrue(type_a_found, "Type A was not found in Response")
        self.assertTrue(type_b_found, "Type B was not found in Response")
        self.assertFalse(type_c_found, "type C was found in Response, but should not be found")



"""
Provide Filters for Pandas based views
"""

import operator
from functools import reduce

from rest_framework.filters import OrderingFilter, SearchFilter

from wbcore import filters as wb_filters
from wbcore.filters import DjangoFilterBackend
from wbcore.filters.utils import check_required_filters


class PandasDjangoFilterBackend(DjangoFilterBackend):
    lookups_operator = {
        "lte": operator.le,
        "lt": operator.lt,
        "gte": operator.ge,
        "gt": operator.gt,
        "exact": operator.eq,
    }

    def filter_queryset(self, request, queryset, view):
        filterset_class = self.get_filterset_class(view, queryset)
        if filterset_class and callable(filterset_class):
            kwargs = self.get_filterset_kwargs(request, queryset, view)
            filterset = filterset_class(**kwargs)
            if filterset.is_valid():
                check_required_filters(request, view, filterset)
        return super().filter_queryset(request, queryset, view)

    def filter_dataframe(self, request, df, view):
        if not df.empty:
            filterset_class = self.get_filterset_class(view, view.get_queryset())
            if filterset_class and callable(filterset_class):
                kwargs = self.get_filterset_kwargs(request, view.get_queryset(), view)
                filter_terms = filterset_class(**kwargs).form.data
                conditions = []
                for filter_term, value in filter_terms.items():
                    if _filter := getattr(filterset_class.Meta, "df_fields", {}).get(filter_term, None):
                        # We support only number for now
                        lookup_expr = getattr(_filter, "lookup_expr", "exact")
                        if isinstance(_filter, wb_filters.NumberFilter):
                            try:
                                conditions.append(
                                    self.lookups_operator[lookup_expr](df[_filter.field_name], float(value))
                                )
                            except ValueError:
                                pass
                if conditions:
                    df = df[reduce(operator.and_, conditions)]
        return df


class PandasSearchFilter(SearchFilter):
    def filter_queryset(self, request, queryset, view):
        return queryset

    def filter_dataframe(self, request, df, view):
        search_fields = self.get_search_fields(view, request)
        search_terms = self.get_search_terms(request)
        search_fields = [field for field in search_fields if field in df.columns]

        if not search_fields or not search_terms or df.empty:
            return df

        conditions = []

        for search_term in search_terms:
            queries = [
                df[field].str.contains(search_term, na=False, case=False, regex=False) for field in search_fields
            ]
            conditions.append(reduce(operator.or_, queries))

        df = df[reduce(operator.and_, conditions)]

        return df


class PandasOrderingFilter(OrderingFilter):
    def get_ordering(self, request, df, view) -> tuple[list[str], list[bool]]:
        """
        Ordering is set by a comma delimited ?ordering=... query parameter.

        The `ordering` query parameter can be overridden by setting
        the `ordering_param` value on the OrderingFilter or by
        specifying an `ORDERING_PARAM` value in the API settings.
        """
        params = request.query_params.get(self.ordering_param)
        base_ordering_fields = view.get_ordering_fields()
        potential_ordering_fields = params.split(",") if params else self.get_default_ordering(view)
        ordering_res = []
        ascending_res = []
        if potential_ordering_fields:
            for ordering_field in potential_ordering_fields:
                ordering_field = ordering_field.strip()
                ascending = True
                if ordering_field.startswith("-"):
                    ordering_field = ordering_field[1:]
                    ascending = False

                if ordering_field in df.columns and ordering_field in base_ordering_fields:
                    ordering_res.append(ordering_field)
                    ascending_res.append(ascending)
        return ordering_res, ascending_res

    def filter_queryset(self, request, queryset, view):
        return queryset

    def filter_dataframe(self, request, df, view):
        ordering_by, ascending_list = self.get_ordering(request, df, view)
        if ordering_by and ascending_list:
            df.sort_values(by=list(ordering_by), ascending=list(ascending_list), inplace=True)
        return df



# Generated by Django 4.2.8 on 2024-01-03 08:58

from contextlib import suppress

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db import migrations


def migrate_content_type(apps, schema_editor):
    DocumentModelRelationship = apps.get_model("documents", "DocumentModelRelationship")

    for model_name in [
        "Classification",
        "Instrument",
        "ClassificationGroup",
        "Deal",
        "Exchange",
        "InstrumentClassificationRelatedInstrument",
        "InstrumentClassificationThroughModel",
        "InstrumentFavoriteGroup",
        "InstrumentList",
        "InstrumentListThroughModel",
        "InstrumentRequest",
        "RelatedInstrumentThroughModel",
        "InstrumentPrice",
    ]:
        with suppress(ContentType.DoesNotExist):
            old_ct = ContentType.objects.get(app_label="wbportfolio", model=model_name.lower())
            new_ct = ContentType.objects.get(app_label="wbfdm", model=model_name.lower())
            DocumentModelRelationship.objects.filter(content_type_id=old_ct.id).update(content_type_id=new_ct.id)


class Migration(migrations.Migration):
    dependencies = (
        [
            ("documents", "0003_alter_documentmodelrelationship_unique_together_and_more"),
            ("wbfdm", "0012_instrumentprice_created_instrumentprice_modified"),
        ]
        if apps.is_installed("wbfdm")
        else [("documents", "0003_alter_documentmodelrelationship_unique_together_and_more")]
    )

    operations = [migrations.RunPython(migrate_content_type)]



import pytest
from django.contrib.contenttypes.models import ContentType
from guardian.shortcuts import UserObjectPermission
from wbcore.contrib.authentication.models.users import Permission
from wbcore.contrib.guardian.models.mixins import PermissionObjectModelMixin
from wbcore.contrib.guardian.utils import (
    assign_permissions,
    get_permission_matrix,
    get_public_user_or_group,
    prune_permissions,
    reload_permissions,
)


@pytest.mark.django_db
class TestAssignPermissionsMap:
    def test_assign_permissions(self, user):
        assign_permissions([("view_user", user, user, True)])

        assert user.has_perm("view_user", user)
        assert not user.has_perm("change_user", user)

    def test_assign_permissions_editable(self, user):
        assign_permissions([("view_user", user, user, True), ("change_user", user, user, False)])

        assert user.has_perm("view_user", user)
        assert user.has_perm("change_user", user)

        assert UserObjectPermission.objects.get(
            user=user,
            content_type_id=ContentType.objects.get_for_model(user).pk,
            object_pk=user.pk,
            permission__codename="view_user",
        ).editable

        assert not UserObjectPermission.objects.get(
            user=user,
            content_type_id=ContentType.objects.get_for_model(user).pk,
            object_pk=user.pk,
            permission__codename="change_user",
        ).editable

    def test_assign_permission_with_non_perm(self, mocker):
        assign_perm = mocker.patch("wbcore.contrib.guardian.utils.assign_perm")
        assign_perm.return_value = None

        assign_permissions([("view_user", mocker.Mock(), mocker.Mock(), True)])

        assign_perm.assert_called_once()


@pytest.mark.django_db
class TestGetPublicUserOrGroup:
    def test_get_public_user_or_group(self, user):
        public_user = get_public_user_or_group()

        assert public_user.count() == 1
        public_user = public_user.first()
        assert public_user
        assert public_user == user
        assert public_user.is_active

    def test_get_public_user_or_group_only_internal(self, internal_user):
        users = get_public_user_or_group(only_internal=True)
        assert users.count() == 1

    def test_get_public_user_or_group_only_internal_with_external_user(self, user):
        users = get_public_user_or_group(only_internal=True)
        assert users.count() == 0


@pytest.mark.django_db
class TestGetPermissionMatrix:
    def test_get_permission_matrix(self, user, mocker):
        def get_permission_for_user(self, created=None):
            return {"view_user": True, "change_user": False}

        queryset = mocker.Mock()
        instance = mocker.Mock()
        queryset.__iter__ = lambda _: iter([instance])
        queryset.all.return_value = iter([instance])
        instance.get_permissions_for_user = get_permission_for_user
        instance.permission_type = PermissionObjectModelMixin.PermissionType.PUBLIC
        instance.PermissionType = PermissionObjectModelMixin.PermissionType

        matrix = get_permission_matrix(queryset, None, None)
        assert list(matrix) == [("view_user", instance, user, True), ("change_user", instance, user, False)]

    def test_get_permission_matrix_with_instance(self, user, mocker):
        def get_permission_for_user(self, created=None):
            return {"view_user": True, "change_user": False}

        queryset = mocker.Mock()
        queryset.filter.return_value = queryset
        instance = mocker.Mock()
        queryset.__iter__ = lambda _: iter([instance])
        queryset.all.return_value = iter([instance])
        instance.get_permissions_for_user = get_permission_for_user
        instance.permission_type = PermissionObjectModelMixin.PermissionType.PUBLIC
        instance.PermissionType = PermissionObjectModelMixin.PermissionType

        matrix = get_permission_matrix(queryset, None, instance)
        next(matrix)
        queryset.filter.assert_called_once_with(id=instance.id)

    def test_get_permission_matrix_private(self, internal_user, mocker):
        def get_permission_for_user(self, created=None):
            return {"view_user": True, "change_user": False}

        queryset = mocker.Mock()
        queryset.filter.return_value = queryset
        instance = mocker.Mock()
        queryset.__iter__ = lambda _: iter([instance])
        queryset.all.return_value = iter([instance])
        instance.get_permissions_for_user = get_permission_for_user
        instance.permission_type = PermissionObjectModelMixin.PermissionType.PRIVATE
        instance.PermissionType = PermissionObjectModelMixin.PermissionType

        matrix = get_permission_matrix(queryset, None, None)
        assert list(matrix) == [
            ("view_user", instance, internal_user, True),
            ("change_user", instance, internal_user, False),
        ]


@pytest.mark.django_db
class TestPrunePermissions:
    def test_prune_permissions(self, user):
        ct = ContentType.objects.get_for_model(user)
        permission = Permission.objects.filter(content_type=ct).first()
        UserObjectPermission.objects.create(
            permission=permission, user=user, content_type=ct, object_pk=user.id, editable=False
        )
        assert UserObjectPermission.objects.filter(permission=permission, content_type=ct, object_pk=user.id).exists()
        prune_permissions(user)
        assert not UserObjectPermission.objects.filter(
            permission=permission, content_type=ct, object_pk=user.id
        ).exists()

    def test_prune_permissions_editable(self, user):
        ct = ContentType.objects.get_for_model(user)
        permission = Permission.objects.filter(content_type=ct).first()
        UserObjectPermission.objects.create(
            permission=permission, user=user, content_type=ct, object_pk=user.id, editable=True
        )
        assert UserObjectPermission.objects.filter(permission=permission, content_type=ct, object_pk=user.id).exists()
        prune_permissions(user)
        assert UserObjectPermission.objects.filter(permission=permission, content_type=ct, object_pk=user.id).exists()

    def test_prune_permissions_editable_with_force(self, user):
        ct = ContentType.objects.get_for_model(user)
        permission = Permission.objects.filter(content_type=ct).first()
        UserObjectPermission.objects.create(
            permission=permission, user=user, content_type=ct, object_pk=user.id, editable=True
        )
        assert UserObjectPermission.objects.filter(permission=permission, content_type=ct, object_pk=user.id).exists()
        prune_permissions(user, True)
        assert not UserObjectPermission.objects.filter(
            permission=permission, content_type=ct, object_pk=user.id
        ).exists()


class TestReloadPermissions:
    def test_reload_permissions(self, mocker):
        get_permission_matrix = mocker.patch("wbcore.contrib.guardian.utils.get_permission_matrix")
        assign_permissions = mocker.patch("wbcore.contrib.guardian.utils.assign_permissions")

        queryset = mocker.Mock()

        reload_permissions(queryset)

        get_permission_matrix.assert_called_once()
        assign_permissions.assert_called_once()

    def test_reload_permissions_with_prune(self, mocker):
        prune_permissions = mocker.patch("wbcore.contrib.guardian.utils.prune_permissions")
        mocker.patch("wbcore.contrib.guardian.utils.get_permission_matrix")
        mocker.patch("wbcore.contrib.guardian.utils.assign_permissions")

        queryset = mocker.Mock()

        reload_permissions(queryset)

        prune_permissions.assert_not_called()

    def test_reload_permissions_with_prune_and_instance(self, mocker):
        prune_permissions = mocker.patch("wbcore.contrib.guardian.utils.prune_permissions")
        mocker.patch("wbcore.contrib.guardian.utils.get_permission_matrix")
        mocker.patch("wbcore.contrib.guardian.utils.assign_permissions")

        queryset = mocker.Mock()
        instance = mocker.Mock()

        reload_permissions(queryset, instance=instance)

        prune_permissions.assert_called_once()



import factory

from wbcore.contrib.documents.models import (
    Document,
    DocumentType,
    ShareableLink,
    ShareableLinkAccess,
)


class DocumentTypeFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("text", max_nb_chars=64)
    parent = None

    class Meta:
        model = DocumentType


class DocumentFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("text", max_nb_chars=64)
    file = factory.django.FileField(data="test")
    document_type = factory.SubFactory(DocumentTypeFactory)
    system_created = False
    system_key = None

    class Meta:
        model = Document


class ShareableLinkFactory(factory.django.DjangoModelFactory):
    document = factory.SubFactory(DocumentFactory)
    manual_invalid = False
    valid_until = None
    one_time_link = False

    class Meta:
        model = ShareableLink


class ShareableLinkAccessFactory(factory.django.DjangoModelFactory):
    shareable_link = factory.SubFactory(ShareableLinkFactory)

    class Meta:
        model = ShareableLinkAccess



from django.db.models import F, QuerySet

from wbcore import viewsets
from wbcore.contrib.example_app.filters import TeamFilter, TeamStadiumFilter
from wbcore.contrib.example_app.models import Team
from wbcore.contrib.example_app.serializers import (
    TeamModelSerializer,
    TeamRepresentationSerializer,
)
from wbcore.contrib.example_app.viewsets.buttons import TeamButtonConfig
from wbcore.contrib.example_app.viewsets.displays import (
    TeamDisplayConfig,
    TeamStadiumDisplayConfig,
)
from wbcore.contrib.example_app.viewsets.endpoints import TeamStadiumEndpointConfig
from wbcore.contrib.example_app.viewsets.titles import (
    TeamStadiumTitleConfig,
    TeamTitleConfig,
)
from wbcore.viewsets.mixins import OrderableMixin


class TeamModelViewSet(OrderableMixin, viewsets.ModelViewSet):
    button_config_class = TeamButtonConfig
    display_config_class = TeamDisplayConfig
    title_config_class = TeamTitleConfig

    queryset = Team.objects.all()
    search_fields = ("name", "city__representation")
    serializer_class = TeamModelSerializer
    ordering_fields = (
        "name",
        "city__representation",
        "founded_date",
        "coach__computed_str",
        "home_stadium__name",
    )
    ordering = ("order", "name")
    filterset_class = TeamFilter

    def get_queryset(self) -> QuerySet[Team]:
        return (
            super()
            .get_queryset()
            .annotate(_group_key=F("id"))
            .select_related("city", "coach", "home_stadium")
            .prefetch_related("opponents")
        )


class TeamStadiumModelViewSet(TeamModelViewSet):
    display_config_class = TeamStadiumDisplayConfig
    ordering_fields = ("name", "city__representation", "founded_date", "coach__computed_str")
    filterset_class = TeamStadiumFilter
    endpoint_config_class = TeamStadiumEndpointConfig
    title_config_class = TeamStadiumTitleConfig

    def get_queryset(self) -> QuerySet[Team]:
        queryset = super().get_queryset()
        if stadium_id := self.kwargs.get("stadium_id"):
            queryset = queryset.filter(home_stadium=stadium_id)
        return queryset


class TeamRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Team.objects.all()
    serializer_class = TeamRepresentationSerializer
    search_fields = ("name",)
    filterset_class = TeamFilter



from .condition import ConditionFilter
from .data import DataFilter
from .process import (
    ProcessFilter,
    ProcessStepFilter,
    ProcessStepProcessFilter,
    ProcessStepStepFilter,
    ProcessWorkflowFilter,
    AssignedProcessStepFilter,
)
from .step import (
    DecisionStepFilter,
    EmailStepFilter,
    FinishStepFilter,
    JoinStepFilter,
    ScriptStepFilter,
    SplitStepFilter,
    StepFilter,
    StepWorkflowFilter,
    UserStepFilter,
    StartStepFilter,
)
from .transition import TransitionFilter
from .workflow import WorkflowFilter



from celery import chain, shared_task
from django.db.models import Model


@shared_task
def save_instance_as_task(instance):
    instance.save(_with_llm=False)


class LLMMixin(Model):
    _llm_config = []

    def get_all_on_save_with_condition(self):
        for config in self._llm_config:
            if config.on_save and config.check_condition(self):
                yield config

    def save(self, _with_llm: bool = True, _llm_synchronous: bool = False, *args, **kwargs):
        super().save(*args, **kwargs)
        if _with_llm:
            tasks = []
            for index, config in enumerate(self.get_all_on_save_with_condition()):
                tasks.append(config.schedule(self, initial=index == 0))
            if tasks:
                res = chain(*tasks, save_instance_as_task.s())
                if _llm_synchronous:
                    res.apply()
                else:
                    res.apply_async()

    class Meta:
        abstract = True



# Generated by Django 5.0.4 on 2024-05-07 07:25

import picklefield.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("io", "0006_alter_exportsource_query_params"),
    ]

    operations = [
        migrations.AlterField(
            model_name="exportsource",
            name="query_params",
            field=picklefield.fields.PickledObjectField(
                blank=True,
                default=list,
                editable=False,
                verbose_name="SQL query parameters to be used with the sql query",
            ),
        ),
    ]



from django.utils.translation import gettext_lazy as _

from wbcore import filters

from .models import ReleaseNote


def get_choices(*args):
    choices = [(rl, rl) for rl in ReleaseNote.objects.all().distinct("module").values_list("module", flat=True)]
    return choices


class ReleaseNoteFilterSet(filters.FilterSet):
    module = filters.ChoiceFilter(label=_("Module"), choices=get_choices)

    read_unread = filters.ChoiceFilter(
        label=_("Read / Unread"),
        choices=(("read", "Read"), ("unread", "Unread")),
        method="get_read_unread",
    )

    def get_read_unread(self, queryset, name, value):
        if value == "read":
            return queryset.filter(user_read=True)
        elif value == "unread":
            return queryset.filter(user_read=False)
        return queryset

    class Meta:
        model = ReleaseNote
        fields = {
            "version": ["icontains"],
            "module": ["exact"],
            "release_date": ["exact", "lte", "gte"],
            "summary": ["icontains"],
        }



from django.db import models

from wbcore.contrib.dataloader.models import Entity
from wbcore.contrib.dataloader.tests.test.dataloaders.proxies import (
    EntityTestDataloaderProxy,
)


class EntityTest(Entity):
    dl_proxy = EntityTestDataloaderProxy
    name = models.CharField(max_length=255)



from wbcore import viewsets
from wbcore.contrib.workflow.filters import DataFilter
from wbcore.contrib.workflow.models import Data
from wbcore.contrib.workflow.serializers import DataModelSerializer
from wbcore.contrib.workflow.viewsets.display import DataDisplayConfig
from wbcore.contrib.workflow.viewsets.endpoints import DataEndpointConfig
from wbcore.contrib.workflow.viewsets.titles import DataTitleConfig


class DataModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/data.md"
    queryset = Data.objects.all()
    serializer_class = DataModelSerializer
    display_config_class = DataDisplayConfig
    title_config_class = DataTitleConfig
    search_fields = ("label",)
    ordering = ("workflow__name", "data_type", "label")
    filterset_class = DataFilter
    endpoint_config_class = DataEndpointConfig
    ordering_fields = (
        "workflow__name",
        "data_type",
        "label",
        "help_text",
        "default",
        "required",
    )

    def get_queryset(self):
        queryset = super().get_queryset().select_related("workflow")
        if workflow_id := self.kwargs.get("workflow_id"):
            return queryset.filter(workflow=workflow_id)
        return queryset



from import_export.widgets import ForeignKeyWidget

from wbcore.contrib.geography.models import Geography


class CountryForeignKeyWidget(ForeignKeyWidget):
    def __init__(self, field="code_3", use_natural_foreign_keys=False, **kwargs):
        super().__init__(Geography, field=field)

    def get_queryset(self, value, row, *args, **kwargs):
        return Geography.countries.all()



from dataclasses import dataclass, field
from typing import Iterable, Optional

from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from rest_framework.request import Request

from wbcore.enums import RequestType
from wbcore.metadata.configs.display.instance_display import Display
from wbcore.serializers import ListSerializer, RepresentationSerializer, Serializer
from wbcore.utils.urls import get_parse_endpoint, get_urlencode_endpoint

from .bases import ButtonConfig, ButtonTypeMixin, ButtonUrlMixin
from .enums import ButtonType, HyperlinkTarget


@dataclass
class DropDownButton(ButtonTypeMixin, ButtonConfig):
    button_type = ButtonType.DROPDOWN
    buttons: Iterable = field(default_factory=tuple)

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()
        self.buttons = tuple(self.buttons)
        assert isinstance(self.buttons, tuple)

    def serialize(self, request, **kwargs):
        res = super().serialize(request, **kwargs)
        res["buttons"] = [
            btn.serialize(request, **kwargs)
            for btn in sorted(self.buttons, key=lambda e: (e.weight, e.label, e.title))
        ]
        return res

    def __hash__(self):
        return hash(self.title)


@dataclass
class WidgetButton(ButtonTypeMixin, ButtonUrlMixin, ButtonConfig):
    button_type = ButtonType.WIDGET
    new_mode: bool = False
    open_new_by_default: bool | None = None

    def serialize(self, request, **kwargs):
        if self.new_mode and self.endpoint:
            endpoint, params = get_parse_endpoint(self.endpoint)
            params["new_mode"] = "true"
            self.endpoint = get_urlencode_endpoint(endpoint, params)
        res = super().serialize(request, **kwargs)
        res["new_mode"] = self.new_mode

        # When open by default is not set but it new mode is True
        # then the default behaviour is to open the content in a new window
        if self.open_new_by_default is not None:
            res["open_new_by_default"] = self.open_new_by_default
        elif self.new_mode:
            res["open_new_by_default"] = True

        return res

    def __hash__(self):
        if self.key:
            return hash(self.key)
        else:
            return hash(self.endpoint)


@dataclass
class HyperlinkButton(ButtonTypeMixin, ButtonUrlMixin, ButtonConfig):
    button_type = ButtonType.HYPERLINK
    target: HyperlinkTarget = HyperlinkTarget.BLANK

    def serialize(self, request, **kwargs):
        res = super().serialize(request, **kwargs)
        res["target"] = self.target.value
        return res

    def __hash__(self):
        if self.key:
            return hash(self.key)
        else:
            return hash(self.endpoint)


@dataclass
class ActionButton(ButtonTypeMixin, ButtonUrlMixin, ButtonConfig):
    button_type = ButtonType.ACTION
    method: RequestType = RequestType.POST
    action_label: str = ""

    description_fields: str = _("<p>Are you sure you want to proceed?</p>")
    instance_display: Optional[Display] = None
    serializer: Optional[type[Serializer]] = None
    confirm_config: ButtonConfig = ButtonConfig(label=gettext_lazy("Confirm"), title=gettext_lazy("Confirm"))
    cancel_config: ButtonConfig = ButtonConfig(
        label=gettext_lazy("Cancel"),
        title=gettext_lazy("Cancel"),
    )

    identifiers: Iterable[str] = field(default_factory=list)

    def __hash__(self):
        if self.key:
            return hash(self.key)
        else:
            return hash(self.endpoint)

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()

    def _get_fields(self, request: Request) -> dict:
        fields = dict()
        rs = RepresentationSerializer
        ls = ListSerializer
        if self.serializer and (field_items := self.serializer(context={"request": request}).fields.items()):
            for name, _field in filter(lambda f: not isinstance(f[1], (rs, ls)), field_items):
                field_name, representation = _field.get_representation(request, name)
                fields[name] = representation

            for name, _field in filter(lambda f: isinstance(f[1], (rs, ls)), field_items):
                field_name, representation = _field.get_representation(request, name)
                fields[representation["related_key"]].update(representation)
                del fields[representation["related_key"]]["related_key"]

        return fields

    def serialize(self, request, **kwargs):
        res = super().serialize(request, **kwargs)
        res["action_label"] = self.action_label
        res["method"] = self.method.value
        res["description_fields"] = self.description_fields
        res["confirm_config"] = self.confirm_config.serialize(request, **kwargs)
        res["cancel_config"] = self.cancel_config.serialize(request, **kwargs)
        res["identifiers"] = self.identifiers
        if self.instance_display:
            res["instance_display"] = self.instance_display.serialize()

        if self.serializer:
            res["fields"] = self._get_fields(request)
        return res



import pytest
from django.test import Client, TestCase
from django.urls import reverse
from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.example_app.factories import (
    PlayerFactory,
    SportPersonFactory,
    TeamFactory,
)
from wbcore.contrib.example_app.models import Player, SportPerson
from wbcore.contrib.example_app.serializers import (
    PlayerModelSerializer,
    SportPersonModelSerializer,
)
from wbcore.contrib.example_app.tests.test_viewsets.test_utils_viewsets import (
    find_instances_in_response,
    get_create_view,
    get_delete_view,
    get_detail_view,
    get_partial_view,
    get_update_view,
)
from wbcore.contrib.example_app.viewsets import (
    PlayerModelViewSet,
    SportPersonModelViewSet,
)


@pytest.mark.django_db
class TestSportPersonModelViewSet(TestCase):
    def setUp(self):
        self.user = SuperUserFactory()
        self.client = Client()
        self.client.force_login(user=self.user)
        self.list_url = reverse("example_app:person-list")
        self.detail_url_str = "example_app:person-detail"

    def test_list_view(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

    def test_create_view(self):
        person = SportPersonFactory()
        response = get_create_view(self.client, person, self.user, self.list_url, SportPersonModelViewSet)
        self.assertEqual(response.status_code, 201)
        self.assertTrue(SportPerson.objects.filter(last_name=person.last_name).exists())

    def test_detail_view(self):
        instance = SportPersonFactory()
        response = get_detail_view(self.client, instance.pk, self.detail_url_str)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["last_name"], instance.last_name)

    def test_update_view(self):
        instance = SportPersonFactory()
        instance.last_name = "Updated Instance"
        instance.profile_image = None
        response = get_update_view(self.client, instance, SportPersonModelSerializer, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["last_name"], instance.last_name)

    def test_partial_update_view(self):
        instance = SportPersonFactory()
        response = get_partial_view(self.client, instance.id, {"last_name": "Updated Instance"}, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["last_name"], instance.last_name)

    def test_delete_view(self):
        instance = SportPersonFactory()
        response = get_delete_view(self.client, self.detail_url_str, instance.pk)
        self.assertEqual(response.status_code, 204)
        self.assertFalse(SportPerson.objects.filter(pk=instance.pk).exists())

    def test_ordering_fields(self):
        person_a = SportPersonFactory(first_name="Hans", last_name="Brecht")
        person_b = SportPersonFactory(first_name="Hans", last_name="Ahrens")
        person_c = SportPersonFactory(first_name="Ralf", last_name="Christ")

        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], SportPerson.objects.count())
        self.assertEqual(response.data["results"][0]["id"], person_b.id)
        self.assertEqual(response.data["results"][1]["id"], person_a.id)
        self.assertEqual(response.data["results"][2]["id"], person_c.id)


@pytest.mark.django_db
class TestPlayerModelViewSet(TestCase):
    def setUp(self):
        self.user = SuperUserFactory()
        self.client = Client()
        self.client.force_login(user=self.user)
        self.detail_url_str = "example_app:player-detail"
        self.list_url = reverse("example_app:player-list")

    def test_list_view(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

    def test_create_view(self):
        player = PlayerFactory()
        response = get_create_view(self.client, player, self.user, self.list_url, PlayerModelViewSet)
        self.assertEqual(response.status_code, 201)
        self.assertTrue(Player.objects.filter(last_name=player.last_name).exists())

    def test_detail_view(self):
        instance = PlayerFactory()
        response = get_detail_view(self.client, instance.pk, self.detail_url_str)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["last_name"], instance.last_name)

    def test_update_view(self):
        instance = PlayerFactory()
        instance.last_name = "Updated Instance"
        response = get_update_view(self.client, instance, PlayerModelSerializer, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["last_name"], instance.last_name)

    def test_partial_update_view(self):
        instance = PlayerFactory()
        response = get_partial_view(self.client, instance.id, {"last_name": "Updated Instance"}, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["last_name"], instance.last_name)

    def test_delete_view(self):
        instance = PlayerFactory()
        response = get_delete_view(self.client, self.detail_url_str, instance.pk)
        self.assertEqual(response.status_code, 204)
        self.assertFalse(Player.objects.filter(pk=instance.pk).exists())

    def test_ordering_fields(self):
        team_a, team_b = TeamFactory(name="Team A"), TeamFactory(name="Team B")
        person_a = PlayerFactory(first_name="Hans", last_name="Ahrens", current_team=team_b)
        person_b = PlayerFactory(first_name="Hans", last_name="Ahrens", current_team=team_a)
        person_c = PlayerFactory(first_name="Ralf", last_name="Brecht")

        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], Player.objects.count())
        self.assertEqual(response.data["results"][0]["id"], person_b.id)
        self.assertEqual(response.data["results"][1]["id"], person_a.id)
        self.assertEqual(response.data["results"][2]["id"], person_c.id)

    def test_player_team(self):
        player_a = PlayerFactory()
        player_b = PlayerFactory(current_team=player_a.current_team)
        player_c = PlayerFactory()
        expected_number_of_player = Player.objects.filter(current_team=player_a.current_team).count()
        player_team_url = reverse("example_app:player-team-list", args=[player_a.current_team.id])
        response = self.client.get(player_team_url)
        player_a_found, player_b_found, player_c_found = find_instances_in_response(
            [player_a, player_b, player_c], response
        )
        self.assertEqual(
            response.data["count"],
            expected_number_of_player,
            f"The answer should contain {expected_number_of_player} players",
        )
        self.assertEqual(response.status_code, 200)
        self.assertTrue(player_a_found, "Player A was not found in Response")
        self.assertTrue(player_b_found, "Player B was not found in Response")
        self.assertFalse(player_c_found, "Player C was found in Response, but should not be found")



# Generated by Django 4.2.6 on 2024-01-31 10:40

import wbcore.contrib.icons.models
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("agenda", "0008_alter_calendaritem_item_type"),
    ]

    operations = [
        migrations.AlterField(
            model_name="calendaritem",
            name="icon",
            field=wbcore.contrib.icons.models.IconField(blank=True, max_length=128, null=True, verbose_name="Icon"),
        ),
    ]



# Generated by Django 4.2.9 on 2024-01-26 15:31

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("workflow", "0003_alter_condition_attribute_name_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userstep",
            name="assignee_method",
            field=models.CharField(
                blank=True,
                choices=[
                    ("manager_of_instance_assignee", "Instance Assignee's Manager"),
                    ("weighted_random", "Weighted Random Group Member"),
                    ("random_group_member", "Random Group Member"),
                ],
                max_length=64,
                null=True,
                verbose_name="Assignee Method",
            ),
        ),
    ]



from wbcore.metadata.configs.display.list_display import (
    ListDisplay,
    TreeGroupLevelOption,
)


def test_tree_group_open_level_list_display():
    ld = dict(
        ListDisplay(
            fields=[],
            tree=True,
            tree_group_field="field",
            tree_group_open_level=3,
            tree_group_level_options=[TreeGroupLevelOption(list_endpoint="endpoint")],
        )
    )
    assert ld["tree_group"]["open_level"] == 3


def test_not_tree_group_open_level_list_display():
    ld = dict(
        ListDisplay(
            fields=[],
            tree=False,
            tree_group_field="field",
            tree_group_open_level=3,
        )
    )
    assert "tree_group" not in ld



import hashlib
import hmac


def get_hash_from_user(user_id: int | str, secret: str) -> str:
    if isinstance(user_id, int):
        user_id = str(user_id)

    return hmac.new(
        bytes(secret, encoding="utf-8"),
        bytes(user_id, encoding="utf-8"),
        digestmod=hashlib.sha256,
    ).hexdigest()



import pytest

from wbcore.contrib.notifications.factories.notification_types import (
    NotificationTypeSettingModelFactory,
)
from wbcore.contrib.notifications.models import NotificationUserToken


@pytest.mark.django_db
class TestNotificationUserToken:
    def test_factory(self, notification_user_token: NotificationUserToken):
        assert isinstance(notification_user_token, NotificationUserToken)
        assert notification_user_token.pk is not None

    def test_to_str(self, notification_user_token: NotificationUserToken):
        assert str(notification_user_token) == "{0.user}: {0.token} ({0.device_type})".format(notification_user_token)

    def test_filter_for_settings(self, notification_user_token: NotificationUserToken):
        setting = NotificationTypeSettingModelFactory(
            user=notification_user_token.user, enable_web=True, enable_mobile=True
        )
        assert setting.enable_web is False
        assert setting.enable_mobile is False
        assert not NotificationUserToken.objects.filter_for_user_settings(setting).exists()

        setting.enable_web = True
        setting.enable_mobile = True
        setting.save()

        assert NotificationUserToken.objects.filter_for_user_settings(setting).first() == notification_user_token  # type: ignore



from contextlib import suppress
from datetime import timedelta

from celery import shared_task
from django.db.models import ProtectedError
from django.utils import timezone
from dynamic_preferences.registries import global_preferences_registry
from tqdm import tqdm

from wbcore.cache.registry import periodic_cache_registry
from wbcore.models import DynamicModel
from wbcore.utils.itertools import get_inheriting_subclasses
from wbcore.utils.models import ComplexToStringMixin, DeleteToDisableMixin


@shared_task
def recompute_latest_modified_dynamic_fields(from_datetime=None, to_datetime=None, timedelta_days=1):
    """
    Daily (expected frequency) recompute of all dynamic model objects that falls within the specified time frame (default to the last 24h).
    For each recently modified object, a subroutine sends a signal to get the dependant object to be recomputed
    Args:
        from_datetime: A datetime, default to today - 24h
        to_datetime: A Datetime, default to today
        timedelta_days: time range, default to 24h
    """
    if not to_datetime:
        to_datetime = timezone.now()
    if not from_datetime:
        from_datetime = timezone.now() - timedelta(days=timedelta_days)

    if from_datetime >= to_datetime:
        raise ValueError("From date needs to be strictly lower than To date")

    # already_updated_objects = set() # We don't pass the list anymore because of in memory issue. For instrument price, we extend 10GB. We migh want to implement a non deterministric approach with Bloom filter for instance

    for subclass in get_inheriting_subclasses(DynamicModel):
        qs = subclass.objects.filter(updated_at__gte=from_datetime, updated_at__lte=to_datetime)
        for already_updated_object in tqdm(qs.distinct().iterator(), total=qs.count()):
            DynamicModel.update_dynamic_fields_of_dependant_objects(already_updated_object)


@shared_task
def recompute_computed_str(debug: bool = False):
    """
    When this task is executed, it will loop over all objects that inherit from ComplexToStringMixin and compare their current computed_str value with the expected one.
    If different, the expected one is saved in place.
    """
    BULK_SIZE = 1000
    for subclass in get_inheriting_subclasses(ComplexToStringMixin):
        if getattr(subclass, "COMPUTED_STR_RECOMPUTE_PERIODICALLY", True):
            objs = []
            if debug:
                qs = tqdm(subclass.objects.iterator(), total=subclass.objects.count())
            else:
                qs = subclass.objects.iterator()
            for instance in qs:
                with suppress(Exception):  # we don't want this task to fail because of
                    new_computed_str = instance.compute_str()
                    if new_computed_str != instance.computed_str:
                        instance.computed_str = new_computed_str
                        objs.append(instance)
                    if len(objs) % BULK_SIZE == 0:
                        subclass.objects.bulk_update(objs, ["computed_str"])
                        objs = []
            if objs:
                subclass.objects.bulk_update(objs, ["computed_str"])


@shared_task
def clean_deleted_objects():
    """
    Periodically arise deleted objects that have passed the retention period from the database.
    """
    retention_period = global_preferences_registry.manager()["wbcore__retention_period"]
    for subclass in get_inheriting_subclasses(DeleteToDisableMixin):
        if subclass.AUTOMATICALLY_CLEAN_SOFT_DELETED_OBJECTS:
            with suppress(ProtectedError):
                subclass.all_objects.filter(
                    is_active=False, deletion_datetime__lte=timezone.now() - timedelta(days=retention_period)
                ).delete()


@shared_task
def refetch_pandas_api_view():
    for cache_api_view in periodic_cache_registry.classes:
        cache_api_view.fetch_cache()



from django.db import connection
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register

from .test.factories import EntityTestFactory


def app_pre_migration(sender, app_config, **kwargs):
    cur = connection.cursor()
    cur.execute("CREATE EXTENSION IF NOT EXISTS pg_trgm;")
    cur.execute("CREATE EXTENSION IF NOT EXISTS btree_gin;")
    cur.execute("CREATE EXTENSION IF NOT EXISTS btree_gist;")


pre_migrate.connect(app_pre_migration)
register(EntityTestFactory)
register(
    EntityTestFactory,
    "entity_test_over_1000",
    dl_parameters={
        "data": {
            "path": "wbcore.contrib.dataloader.tests.test.dataloaders.dataloaders.RandomDataOver1000",
            "parameters": {},
        }
    },
)



import pytest
from django.db.models import F
from django.db.models.expressions import OrderBy

from wbcore.viewsets.mixins import map_ordering


class TestWBCoreOrderingFilter:
    @pytest.mark.parametrize(
        ("input", "output"),
        [
            ("test_str", OrderBy(F("test_str"))),
            ("-test_str", OrderBy(F("test_str"), descending=True)),
            ("test_str__nulls_last", OrderBy(F("test_str"), nulls_last=True)),
            ("-test_str__nulls_last", OrderBy(F("test_str"), nulls_last=True, descending=True)),
            ("test_str__nulls_first", OrderBy(F("test_str"), nulls_first=True)),
            ("-test_str__nulls_first", OrderBy(F("test_str"), nulls_first=True, descending=True)),
        ],
    )
    def test_map_ordering(self, input, output):
        assert map_ordering(input) == output



from wbcore import serializers

from .models import ReleaseNote


class ReleaseNoteModelSerializer(serializers.ModelSerializer):
    user_read = serializers.BooleanField()
    user_read_icon = serializers.IconSelectField()

    class Meta:
        model = ReleaseNote
        fields = ("id", "version", "release_date", "module", "summary", "notes", "user_read", "user_read_icon")



from typing import Any

from rest_framework.metadata import BaseMetadata
from rest_framework.request import Request
from rest_framework.views import View

from wbcore.metadata.configs.base import WBCoreViewConfig


class WBCoreMetadata(BaseMetadata):
    """A metadata class for WBCore application views."""

    def determine_metadata(self, request: Request, view: View) -> dict[str, Any]:
        """
        Returns a dictionary containing metadata for the given request and view.

        This method iterates over all the WBCoreViewConfig instances associated
        with the given view and adds their metadata to the returned dictionary.

        Args:
            request (Request): An HTTP request instance.
            view (View): A Django Rest Framework view.

        Returns:
            A dictionary containing metadata for the given request and view.
        """

        metadata = dict()

        for _, config_class in WBCoreViewConfig.get_all_configs(view):
            config = config_class.for_view(view, request)
            metadata[config.metadata_key] = config._get_metadata()

        return metadata



import json

from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import JsonLexer


class PrintColor:
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


def print_bold(text, color=None):
    if color:
        print(f"{PrintColor.BOLD}{color}{text}{PrintColor.END}")  # noqa: T201
    else:
        print(f"{PrintColor.BOLD}{text}{PrintColor.END}")  # noqa: T201


def print_dict(dictionary):
    print(highlight(json.dumps(dictionary, indent=4), JsonLexer(), TerminalFormatter()))  # noqa: T201



# Generated by Django 4.1.8 on 2023-04-28 13:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcore", "0001_initial_squashed_squashed_0010_preset_appliedpreset"),
    ]

    operations = [
        migrations.CreateModel(
            name="GenericModel",
            fields=[
                ("id", models.IntegerField(primary_key=True, serialize=False)),
                ("label", models.CharField(max_length=256)),
            ],
            options={
                "managed": False,
            },
        ),
    ]



from .contacts import (
    AddressContact,
    BankingContact,
    ContactLocationChoices,
    ContactTelephoneChoices,
    EmailContact,
    SocialMediaContact,
    TelephoneContact,
    WebsiteContact,
)
from .entries import (
    BankCompanyManager,
    Company,
    CompanyType,
    CustomerStatus,
    Entry,
    Person,
    PersonEmployeeManager,
    Specialization,
)
from .relationships import (
    ClientManagerRelationship,
    EmployerEmployeeRelationship,
    Position,
    Relationship,
    RelationshipType,
)



import logging

from django.conf import settings
from django.contrib.auth.models import Group, Permission
from django.core.mail import EmailMultiAlternatives
from django.db import IntegrityError
from django.db.models import Q
from django.db.models.functions import Coalesce
from django.db.utils import DataError
from django.shortcuts import get_object_or_404, redirect, render
from django.template.loader import render_to_string
from django.urls import reverse_lazy
from django.utils.http import urlencode
from django.utils.translation import gettext as _
from rest_framework import filters, permissions, status
from rest_framework.decorators import (
    action,
    api_view,
    authentication_classes,
    permission_classes,
    renderer_classes,
)
from rest_framework.permissions import AllowAny
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework_simplejwt.views import TokenObtainPairView, TokenRefreshView

from wbcore import viewsets
from wbcore.utils.html import convert_html2text

from ..models import User, UserActivity
from ..serializers import (
    ChangePasswordSerializer,
    GroupModelSerializer,
    GroupRepresentationSerializer,
    MyTokenObtainPairSerializer,
    MyTokenRefreshSerializer,
    PermissionModelSerializer,
    PermissionRepresentationSerializer,
    UserModelSerializer,
    UserProfileModelSerializer,
    UserRepresentationSerializer,
)
from .buttons import UserModelButtonConfig, UserProfileButtonConfig
from .display import (
    UserModelDisplay,
    UserPermissionModelDisplay,
    UserProfileModelDisplay,
)
from .endpoints import (
    UserPermissionsModelEndpointConfig,
    UserProfileModelEndpointConfig,
)
from .titles import (
    UserModelTitleConfig,
    UserPermissionsModelTitleConfig,
    UserProfileModelTitleConfig,
)

logger = logging.getLogger()


@api_view(("POST",))
@renderer_classes((JSONRenderer,))
@permission_classes([AllowAny])
@authentication_classes([])
def reset_password_email(request):
    try:
        user = User.objects.get(email=request.data["email"])
        user.reset_password(request)
    except Exception:
        pass
    return Response(
        {
            "status": "ok",
            "msg": _("If the email matches a user, it will receive an email inviting him to reset his password."),
        },
        status=status.HTTP_200_OK,
    )


@api_view(("POST",))
@renderer_classes((JSONRenderer,))
@permission_classes([AllowAny])
@authentication_classes([])
def register_user(request):
    email = request.POST.get("email", "")
    first_name = request.POST.get("first_name", "")
    last_name = request.POST.get("last_name", "")
    password = request.POST.get("password", "")
    if len(first_name) > 100 or len(last_name) > 100:
        return Response(
            data={"status": "fail", "msg": _("first and last name are too long. please provider a shorter name.")},
            status=status.HTTP_400_BAD_REQUEST,
        )
    if len(password) > 128:
        return Response(
            data={
                "status": "fail",
                "msg": _("password is too long. please provider a password shorter than 128 characters."),
            },
            status=status.HTTP_400_BAD_REQUEST,
        )
    if len(email) > 255:
        return Response(
            data={
                "status": "fail",
                "msg": _("email is too long. please provider an email shorter than 255 characters."),
            },
            status=status.HTTP_400_BAD_REQUEST,
        )
    if email and first_name and last_name and password:
        try:
            user = User.create_with_attributes(email, password, first_name=first_name, last_name=last_name)
            query_params = {}
            if success_redirect_url := request.POST.get("success_redirect_url", None):
                query_params["success_redirect_url"] = success_redirect_url
            if error_redirect_url := request.POST.get("error_redirect_url", None):
                query_params["error_redirect_url"] = error_redirect_url

            token = user.generate_temporary_token()
            url = f'{reverse("wbcore:authentication:activate", args=[user.uuid, token], request=request)}?{urlencode(query_params)}'

            # Construct registration mail and send
            rendered_message = render_to_string("user_registration_email.html", {"user": user, "url": url})
            email = EmailMultiAlternatives(
                _("Activate your account."),
                convert_html2text(rendered_message),
                settings.DEFAULT_FROM_EMAIL,
                to=[email],
            )
            email.attach_alternative(rendered_message, "text/html")
            email.send()
            return Response(
                {"status": "success", "msg": _("Please confirm your email address to complete the registration")},
                status=status.HTTP_200_OK,
            )
        except IntegrityError:
            return Response(
                {"status": "fail", "msg": _("Your account already exists")}, status=status.HTTP_409_CONFLICT
            )
        except DataError:
            return Response(
                {"status": "fail", "msg": _("Something went wrong with the submitted data, please try again later")},
                status=status.HTTP_400_BAD_REQUEST,
            )
    return Response(
        {"status": "fail", "msg": _("email, first_name, last_name and password must be provided")},
        status=status.HTTP_400_BAD_REQUEST,
    )


@api_view(("GET",))
@permission_classes([AllowAny])
@authentication_classes([])
def activate_user(request, uuid, token):
    user = get_object_or_404(User, uuid=uuid)
    if user.verify_temporary_token(token):
        user.is_active = True
        user.is_register = True
        user.save()
        # return redirect('home')
        if success_redirect_url := request.GET.get("success_redirect_url", None):
            return redirect(success_redirect_url, status=status.HTTP_200_OK)
        return render(
            request,
            "activate_confirm.html",
            {
                "title": _("Activation Successful"),
                "message": _("You are successfully registered. Please login"),
                "button_title": _("Login"),
                "redirect_url": getattr(settings, "AUTHENTICATION_ACTIVATION_REDIRECTION_URL", "/"),
            },
            status=status.HTTP_200_OK,
        )
    if error_redirect_url := request.GET.get("error_redirect_url", None):
        return redirect(error_redirect_url)
    else:
        return render(
            request,
            "activate_confirm.html",
            {
                "title": _("Activation Failed"),
                "message": _(
                    "The registration link you clicked was probably expired or your account is already activated. Please try to register again."
                ),
                "button_title": _("Register"),
                "redirect_url": reverse_lazy("wbcore:authentication:register"),
            },
            status=status.HTTP_400_BAD_REQUEST,
        )


class UserRepresentationViewSet(viewsets.RepresentationViewSet):
    filter_backends = (filters.OrderingFilter, filters.SearchFilter)
    queryset = User.objects.none()
    serializer_class = UserRepresentationSerializer

    def get_queryset(self):
        if self.request.user.is_superuser or self.request.user.profile.is_internal:
            return User.objects.all()
        return User.objects.filter(id=self.request.user.id)

    ordering_fields = ("username", "email", "date_joined")
    search_fields = ("profile__first_name", "profile__last_name", "email")
    ordering = ["email"]


class GroupRepresentationViewSet(viewsets.RepresentationViewSet):
    filter_backends = (
        filters.OrderingFilter,
        filters.SearchFilter,
    )

    ordering_fields = ordering = ("name",)
    search_fields = ("name",)
    queryset = Group.objects.all()
    serializer_class = GroupRepresentationSerializer
    ordering = ["name"]


class PermissionRepresentationViewSet(viewsets.RepresentationViewSet):
    filter_backends = (
        filters.OrderingFilter,
        filters.SearchFilter,
    )

    ordering_fields = ordering = ("name",)
    search_fields = ("name",)
    queryset = Permission.objects.all()
    serializer_class = PermissionRepresentationSerializer


class GroupViewSet(viewsets.ModelViewSet):
    queryset = Group.objects.all()
    serializer_class = GroupModelSerializer


class PermissionViewSet(viewsets.ModelViewSet):
    queryset = Permission.objects.all()
    serializer_class = PermissionModelSerializer


class UserPermissionsModelViewSet(PermissionViewSet):
    title_config_class = UserPermissionsModelTitleConfig
    endpoint_config_class = UserPermissionsModelEndpointConfig
    display_config_class = UserPermissionModelDisplay

    def get_queryset(self):
        user = User.objects.get(id=self.kwargs["user_id"])
        return Permission.objects.filter(Q(user__id=user.id) | Q(group__in=user.groups.all())).distinct()


class UserProfileModelViewSet(viewsets.ModelViewSet):
    title_config_class = UserProfileModelTitleConfig
    endpoint_config_class = UserProfileModelEndpointConfig
    button_config_class = UserProfileButtonConfig
    display_config_class = UserProfileModelDisplay

    serializer_class = UserProfileModelSerializer
    queryset = User.objects.all()

    def get_queryset(self):
        return User.objects.filter(id=self.request.user.id)

    @action(detail=True, methods=["PATCH"], permission_classes=[permissions.IsAuthenticated])
    def change_password(self, request, pk=None):
        user = get_object_or_404(User, id=pk)
        data = request.POST.dict()
        data["user"] = user
        serializer = ChangePasswordSerializer(data=data, context={"request": request})
        if serializer.is_valid():
            user.set_password(request.data["new_password"])
            user.save()
            return Response({"__notification": {"title": _("Password changed.")}}, status=status.HTTP_200_OK)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["GET"], permission_classes=[permissions.IsAuthenticated])
    def reset_settings(self, request, pk=None):
        return Response(True)


class UserModelViewSet(viewsets.ModelViewSet):
    queryset = User.objects.all()
    search_fields = ("profile__computed_str", "email", "profile__emails__address", "username")
    ordering_fields = ordering = ("email", "date_joined", "last_connection")

    serializer_class = UserModelSerializer

    filterset_fields = {
        "email": ["exact", "icontains"],
        "is_active": ["exact"],
        "date_joined": ["gte", "exact", "lte"],
    }
    title_config_class = UserModelTitleConfig
    button_config_class = UserModelButtonConfig
    display_config_class = UserModelDisplay

    def get_queryset(self):
        if self.request.user.has_perm("authentication.administrate_user"):
            qs = User.objects.all()
        elif self.request.user.profile.is_internal:
            qs = User.objects.filter(profile__relationship_managers=self.request.user.profile)
        else:
            qs = User.objects.filter(id=self.request.user.id)

        return (
            qs.annotate(last_connection=Coalesce(UserActivity.get_latest_login_datetime_subquery(use_user=True), None))
            .select_related("profile")
            .prefetch_related("groups", "user_permissions")
        )

    @action(detail=True, methods=["GET"], permission_classes=[permissions.IsAuthenticated])
    def reset_password(self, request, pk=None):
        user = get_object_or_404(User, id=pk)
        try:
            user.reset_password(request)
            return Response({"detail": _("Reset Password E-Mail sent: Check your mailbox")})
        except Exception:
            return Response({"detail": _("Reset Password Failed. Please contact an administrator")}, status=500)


class MyTokenObtainPairModelView(TokenObtainPairView):
    serializer_class = MyTokenObtainPairSerializer


class MyTokenRefreshModelView(TokenRefreshView):
    serializer_class = MyTokenRefreshSerializer



import pytest
from django.db import IntegrityError
from wbcore.contrib.example_app.factories import (
    EventFactory,
    EventTypeFactory,
    PlayerFactory,
)


@pytest.mark.django_db
class TestEvent:
    def test_str(self):
        event = EventFactory(
            minute=69,
            person__first_name="Test",
            person__last_name="Person",
            event_type__name="Test Event Type",
            match__sport__match_duration=90,
        )
        assert event.__str__() == "Test Event Type (69.) - Test Person"

    @pytest.mark.parametrize(
        "home, away, points, home_score, away_score", [(2, 1, 5, 7, 1), (0, 3, 2, 2, 3), (1, 1, 0, 1, 1)]
    )
    def test_save_home_scored(self, home, away, points, home_score, away_score):
        person = PlayerFactory()
        event = EventFactory(
            person=person,
            match__home=person.current_team,
            match__score_home=home,
            match__score_away=away,
            event_type__points=points,
        )
        assert event.match.score_home == home_score
        assert event.match.score_away == away_score

    @pytest.mark.parametrize(
        "home, away, points, home_score, away_score", [(41, 27, 4, 41, 31), (19, 0, 2, 19, 2), (22, 23, 0, 22, 23)]
    )
    def test_save_away_scored(self, home, away, points, home_score, away_score):
        person = PlayerFactory()
        event = EventFactory(
            person=person,
            match__away=person.current_team,
            match__score_home=home,
            match__score_away=away,
            event_type__points=points,
        )
        assert event.match.score_home == home_score
        assert event.match.score_away == away_score

    def test_save_no_team(self):
        person = PlayerFactory()
        event = EventFactory(
            person=person,
            match__score_home=2,
            match__score_away=1,
            event_type__points=4,
        )
        assert event.match.score_home == 2
        assert event.match.score_away == 1

    def test_save_no_player(self):
        event = EventFactory(
            match__score_home=2,
            match__score_away=1,
            event_type__points=4,
        )
        assert event.match.score_home == 2
        assert event.match.score_away == 1

    def test_unique_constraint(self):
        event = EventFactory()
        with pytest.raises(IntegrityError):
            EventFactory(person=event.person, minute=event.minute, event_type=event.event_type, match=event.match)


@pytest.mark.django_db
class TestEventType:
    def test_str(self):
        event_type = EventTypeFactory(name="Test Event Type")
        assert event_type.__str__() == "Test Event Type"

    def test_unique_constraint(self):
        event_type = EventTypeFactory()
        with pytest.raises(IntegrityError):
            EventTypeFactory(name=event_type.name, sport=event_type.sport)



from django.db import models


class TagModelMixin(models.Model):
    tag_detail_endpoint = models.CharField(max_length=255, null=True, blank=True)
    tag_representation = models.CharField(max_length=512, null=True, blank=True)
    tags = models.ManyToManyField(
        to="tags.Tag",
        related_name="%(app_label)s_%(class)s_items",
        blank=True,
    )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.__class__.objects.filter(id=self.id).update(
            tag_detail_endpoint=self.get_tag_detail_endpoint(),
            tag_representation=self.get_tag_representation(),
        )

    def get_tag_detail_endpoint(self):
        raise NotImplementedError("When using Tags, you must implement get_tag_detail_endpoint")

    def get_tag_representation(self):
        raise NotImplementedError("When using Tags, you must implement get_tag_representation")

    class Meta:
        abstract = True



from configurations import values
from django.conf import settings


class Maintenance:
    MAINTENANCE_MODE_STR: str | None = values.Value("False", environ_prefix=None)

    # We allow maintenance mode to be either True, False or None. In case of None, the specified Backend will take over.
    def MAINTENANCE_MODE(self) -> bool | None:
        if value := self.MAINTENANCE_MODE_STR:
            normalized_value = value.strip().lower()
            if normalized_value in values.BooleanValue.true_values:
                return True
            elif normalized_value in values.BooleanValue.false_values:
                return False
        return None

    MAINTENANCE_MODE_STATE_BACKEND = values.Value(
        "maintenance_mode.backends.StaticStorageBackend", environ_prefix=None
    )
    MAINTENANCE_MODE_STATE_BACKEND_FALLBACK_VALUE = False
    MAINTENANCE_MODE_STATE_FILE_NAME = values.Value("maintenance_mode_state.txt", environ_prefix=None)

    MAINTENANCE_MODE_IGNORE_ADMIN_SITE = values.BooleanValue(True, environ_prefix=None)
    MAINTENANCE_MODE_IGNORE_ANONYMOUS_USER = False
    MAINTENANCE_MODE_IGNORE_AUTHENTICATED_USER = False
    MAINTENANCE_MODE_IGNORE_STAFF = values.BooleanValue(False, environ_prefix=None)
    MAINTENANCE_MODE_IGNORE_SUPERUSER = values.BooleanValue(True, environ_prefix=None)
    MAINTENANCE_MODE_IGNORE_IP_ADDRESSES = values.ListValue([], environ_prefix=None)

    MAINTENANCE_MODE_GET_CONTEXT = None
    MAINTENANCE_MODE_IGNORE_URLS = values.ListValue([], environ_prefix=None)

    MAINTENANCE_MODE_LOGOUT_AUTHENTICATED_USER = False
    MAINTENANCE_MODE_REDIRECT_URL = values.Value(None, environ_prefix=None)
    MAINTENANCE_MODE_RESPONSE_TYPE = values.Value("html", environ_prefix=None)
    MAINTENANCE_MODE_TEMPLATE = values.Value("errors/503.html", environ_prefix=None)

    MAINTENANCE_MODE_STATUS_CODE = values.IntegerValue(503, environ_prefix=None)
    MAINTENANCE_MODE_RETRY_AFTER = values.IntegerValue(3600, environ_prefix=None)  # 1 hour

    MAINTENANCE_MODE_IGNORE_TESTS = False
    MAINTENANCE_MODE_GET_CONTEXT = "wbcore.configurations.configurations.maintenance.get_context"

    MAINTENANCE_MODE_HEADER = values.Value("We're doing some maintenance", environ_prefix=None)
    MAINTENANCE_MODE_DESCRIPTION = values.Value(
        "Our site is currently undergoing scheduled maintenance and upgrades, but will return shortly. Thank for your patience",
        environ_prefix=None,
    )


def get_context(request=None):
    return {"header": settings.MAINTENANCE_MODE_HEADER, "description": settings.MAINTENANCE_MODE_DESCRIPTION}



# Generated by Django 5.0.6 on 2024-06-19 06:28

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0008_bankingcontact_access"),
    ]

    operations = [
        migrations.RenameField(
            model_name="entry",
            old_name="external_identfier",
            new_name="external_identifier",
        ),
        migrations.AlterField(
            model_name="entry",
            name="external_identifier",
            field=models.CharField(blank=True, max_length=128, null=True, unique=True),
        ),
    ]



from django.apps import AppConfig


class TagsConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "wbcore.contrib.tags"

    def ready(self):
        import wbcore.contrib.tags.signals  # noqa: F401



from django.conf import settings
from rest_framework.reverse import reverse

from wbcore.metadata.configs.endpoints import EndpointViewConfig


class CalendarItemEndpointConfig(EndpointViewConfig):
    def _get_instance_endpoint(self, **kwargs):
        return "{{endpoint}}"

    def get_create_endpoint(self, **kwargs):
        return f"{reverse(settings.DEFAULT_CREATE_ENDPOINT_BASENAME, args=[], request=self.request)}?participants={self.request.user.profile.id}&new_mode=True"

    def get_delete_endpoint(self, **kwargs):
        try:
            if self.view.get_object().can_delete(self.request.user):
                return super().get_delete_endpoint(**kwargs)
        except (AssertionError, AttributeError, KeyError):
            return None



# Generated by Django 4.1.8 on 2023-04-06 06:43

import django.contrib.postgres.fields
import django.contrib.postgres.indexes
import django.contrib.postgres.search
import django.db.models.deletion
import mptt.fields
import timezone_field.fields
from django.contrib.postgres.operations import BtreeGinExtension, TrigramExtension
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        TrigramExtension(),
        BtreeGinExtension(),
        migrations.CreateModel(
            name="Geography",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("name", models.CharField(max_length=255)),
                ("short_name", models.CharField(max_length=255)),
                (
                    "alternative_names",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=255),
                        blank=True,
                        default=list,
                        help_text="Alternative Names",
                        size=None,
                        verbose_name="Alternative Names",
                    ),
                ),
                ("representation", models.CharField(blank=True, default="", max_length=255)),
                ("search_vector", django.contrib.postgres.search.SearchVectorField(null=True)),
                ("trigram_search_vector", models.CharField(blank=True, max_length=1024, null=True)),
                (
                    "code_2",
                    models.CharField(
                        blank=True, max_length=2, null=True, verbose_name="2 Character Alphanumeric Code"
                    ),
                ),
                (
                    "code_3",
                    models.CharField(
                        blank=True, max_length=3, null=True, verbose_name="3 Character Alphanumeric Code"
                    ),
                ),
                ("code", models.CharField(blank=True, max_length=16, null=True, verbose_name="Code")),
                ("population", models.PositiveIntegerField(blank=True, null=True)),
                ("ranking", models.PositiveIntegerField(default=1)),
                (
                    "time_zone",
                    timezone_field.fields.TimeZoneField(blank=True, choices_display="WITH_GMT_OFFSET", null=True),
                ),
                ("lft", models.PositiveIntegerField(editable=False)),
                ("rght", models.PositiveIntegerField(editable=False)),
                ("tree_id", models.PositiveIntegerField(db_index=True, editable=False)),
                ("level", models.PositiveIntegerField(editable=False)),
                (
                    "parent",
                    mptt.fields.TreeForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="children",
                        to="geography.geography",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Geographies",
            },
        ),
        migrations.AddIndex(
            model_name="geography",
            index=django.contrib.postgres.indexes.GinIndex(fields=["search_vector"], name="geography_sv_gin_idx"),
        ),
        migrations.AddIndex(
            model_name="geography",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["trigram_search_vector"],
                name="geography_trigram_sv_gin_idx",
                opclasses=["gin_trgm_ops"],
            ),
        ),
        migrations.AddConstraint(
            model_name="geography",
            constraint=models.UniqueConstraint(
                condition=models.Q(("level", 1)),
                fields=("level", "code_2"),
                name="geography_geography_level_code2__uniq",
            ),
        ),
        migrations.AddConstraint(
            model_name="geography",
            constraint=models.UniqueConstraint(
                condition=models.Q(("level", 1)),
                fields=("level", "code_3"),
                name="geography_geography_level_code3__uniq",
            ),
        ),
    ]



from .base import WBCoreViewConfig


class IdentifierViewConfig(WBCoreViewConfig):
    metadata_key = "identifier"
    config_class_attribute = "identifier_config_class"

    # TODO: This does not yet work on the frontend, but should
    def get_metadata(self) -> str | None:
        if identifier := getattr(self.view, "IDENTIFIER", None):
            return identifier

        content_type = self.view.get_content_type()  # type: ignore
        identifier = f"{content_type.app_label}:{content_type.model}"

        return identifier


class DependantIdentifierViewConfig(WBCoreViewConfig):
    metadata_key = "dependant_identifier"
    config_class_attribute = "dependant_identifier_config_class"

    # TODO: This should return a list, shouldn't it?
    def get_metadata(self) -> str | None:
        return getattr(self.view, "DEPENDANT_IDENTIFIER", None)



# Generated by Django 4.1.8 on 2023-04-14 14:12

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("agenda", "0002_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="calendaritem",
            name="endpoint_basename",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name="calendaritem",
            name="item_type",
            field=models.CharField(
                choices=[("AbsenceRequest", "Absence Request"), ("Activity", "Activity"), ("Task", "Task")],
                default="Activity",
                max_length=255,
                verbose_name="Type",
            ),
        ),
        migrations.AlterField(
            model_name="calendaritem",
            name="item_type",
            field=models.CharField(
                choices=[
                    ("CalendarItem", "Calendar Item"),
                    ("AbsenceRequest", "Absence Request"),
                    ("Activity", "Activity"),
                    ("Task", "Task"),
                ],
                default="CalendarItem",
                max_length=255,
                verbose_name="Type",
            ),
        ),
    ]



from .contacts import (
    AddressContactCompanyFilter,
    AddressContactPersonFilter,
    BankingContactCompanyFilter,
    BankingContactFilter,
    BankingContactPersonFilter,
    EmailContactCompanyFilter,
    EmailContactPersonFilter,
    SocialMediaContactCompanyFilter,
    SocialMediaContactPersonFilter,
    TelephoneContactCompanyFilter,
    TelephoneContactPersonFilter,
    WebsiteContactCompanyFilter,
    WebsiteContactPersonFilter,
)
from .entries import CompanyFilter, EntryFilter, PersonFilter, UserIsManagerFilter
from .relationships import (
    ClientManagerFilter,
    RelationshipEntryFilter,
    RelationshipFilter,
    RelationshipTypeFilter,
)



import pytest
from pytest_mock import MockerFixture

from wbcore.contrib.authentication.models import Group, User
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.workflow.models import Process, ProcessStep, Step
from wbcore.contrib.workflow.workflows import (
    manager_of_instance_assignee,
    random_group_member,
    weighted_random,
)


class TestWorkflowAssignees:
    @pytest.fixture
    def mocked_active_user(self, mocker: MockerFixture):
        return mocker.MagicMock(spec=User, is_active=True)

    @pytest.fixture
    def mocked_group(self, mocker: MockerFixture):
        return mocker.MagicMock(spec=Group)

    def test_manager_of_instance_assignee_person(self, mocker: MockerFixture, mocked_process_step, mocked_active_user):
        # Arrange
        kwargs = {"assignee_field": "profile", "assignee_type": "entry"}

        user_a = mocker.MagicMock(spec=User, is_active=False)
        manager_a = mocker.MagicMock(spec=Entry, user_account=user_a)
        manager_b = mocker.MagicMock(spec=Entry, user_account=mocked_active_user)

        assignee_mock = mocker.MagicMock()

        mocked_process_step.process.instance = mocker.MagicMock(spec=User)
        mocked_process_step.process.instance.profile = assignee_mock

        assignee_mock.relationship_managers.all.return_value = [manager_a, manager_b]
        # Act
        result = manager_of_instance_assignee(mocked_process_step, **kwargs)
        # Assert
        assert result == manager_b.user_account

    def test_manager_of_instance_assignee_no_kwargs(self, mocked_process_step, mocker: MockerFixture):
        mocked_process_step.step.get_casted_step.return_value = mocker.MagicMock(spec=Step)
        result = manager_of_instance_assignee(mocked_process_step)
        assert result is None
        mocked_process_step.step.get_casted_step().set_failed.assert_called_once_with(
            mocked_process_step,
            "Error in assignee method: Incorrect input in kwargs field!",
        )

    def test_manager_of_instance_assignee_no_assignee_field(self, mocked_process_step):
        kwargs = {"test": "profile", "assignee_type": "entry"}
        result = manager_of_instance_assignee(mocked_process_step, **kwargs)
        assert result is None
        mocked_process_step.step.get_casted_step().set_failed.assert_called_once_with(
            mocked_process_step,
            "Error in assignee method: Incorrect input in kwargs field!",
        )

    def test_manager_of_instance_assignee_no_assignee_type(self, mocked_process_step):
        kwargs = {"assignee_field": "profile", "test": "entry"}
        result = manager_of_instance_assignee(mocked_process_step, **kwargs)
        assert result is None
        mocked_process_step.step.get_casted_step().set_failed.assert_called_once_with(
            mocked_process_step,
            "Error in assignee method: Incorrect input in kwargs field!",
        )

    def test_manager_of_instance_assignee_wrong_assignee_field(self, mocked_process_step):
        kwargs = {"assignee_field": "profile", "assignee_type": "entry"}
        result = manager_of_instance_assignee(mocked_process_step, **kwargs)
        assert result is None
        mocked_process_step.step.get_casted_step().set_failed.assert_called_once_with(
            mocked_process_step,
            "Error in assignee method: Incorrect input in kwargs field!",
        )

    def test_manager_of_instance_assignee_wrong_assignee_type(self, mocked_process_step, mocked_active_user):
        kwargs = {"assignee_field": "profile", "assignee_type": "test"}
        mocked_process_step.process.instance = mocked_active_user
        result = manager_of_instance_assignee(mocked_process_step, **kwargs)
        assert result is None
        mocked_process_step.step.get_casted_step().set_failed.assert_called_once_with(
            mocked_process_step,
            "Error in assignee method: Incorrect input in kwargs field!",
        )

    def test_weighted_random_with_prior_assignments(self, mocker: MockerFixture, mocked_process_step, mocked_group):
        # Arrange
        user_a = mocker.MagicMock(spec=User, pk=1)
        user_b = mocker.MagicMock(spec=User, pk=2)
        mock_group_users = [user_a, user_b]

        mocked_group.user_set.all.return_value = mock_group_users
        mocked_group.user_set.count.return_value = len(mock_group_users)

        mocked_process_step.group = mocked_group
        mocked_process_step.process = mocker.MagicMock(spec=Process)

        mocked_choices = mocker.patch("wbcore.contrib.workflow.workflows.assignees.choices", return_value=[user_a])
        mocked_qs = mocker.patch("wbcore.contrib.workflow.workflows.assignees.ProcessStep.objects.filter")

        mocked_query = mocked_qs.return_value
        mocked_query.exclude.return_value.values_list.return_value = [1, 2, 2]
        # Act
        result = weighted_random(mocked_process_step)

        mocked_choices.assert_called_once_with(mock_group_users, weights=mocker.ANY)
        assignees, weights = (
            mocked_choices.call_args.args[0],
            mocked_choices.call_args.kwargs["weights"],
        )
        # Assert
        assert assignees == mock_group_users
        assert weights == pytest.approx([2 / 3, 1 / 3], rel=1e-2)
        assert result in mock_group_users

    def test_weighted_random_no_group(self, mocker: MockerFixture, mocked_process_step):
        # Arrange
        mocked_process_step.group = None
        mocked_failed_method = mocker.patch.object(mocked_process_step.step.get_casted_step(), "set_failed")
        # Act
        result = weighted_random(mocked_process_step)
        # Assert
        mocked_failed_method.assert_called_once()
        assert result is None

    def test_weighted_random_empty_group(self, mocker: MockerFixture, mocked_process_step, mocked_group):
        # Arrange
        mocked_group.user_set.count.return_value = 0
        mocked_process_step.group = mocked_group
        mocked_fail_method = mocker.patch.object(mocked_process_step.step.get_casted_step(), "set_failed")
        # Act
        result = weighted_random(mocked_process_step)
        # Assert
        mocked_fail_method.assert_called_once()
        assert result is None

    def test_weighted_random_no_prior_assignments(self, mocker: MockerFixture, mocked_process_step, mocked_group):
        # Arrange
        user_a = mocker.MagicMock(spec=User, pk=1)
        user_b = mocker.MagicMock(spec=User, pk=2)

        mock_group_users = [user_a, user_b]
        mocked_group.user_set.all.return_value = mock_group_users = [user_a, user_b]
        mocked_group.user_set.count.return_value = 2
        mocked_process_step.group = mocked_group

        mocked_choices = mocker.patch(
            "wbcore.contrib.workflow.workflows.assignees.choices",
            return_value=[user_b],
        )
        mocked_qs = mocker.patch("wbcore.contrib.workflow.workflows.assignees.ProcessStep.objects.filter")
        mocked_query = mocked_qs.return_value
        mocked_query.exclude.return_value.values_list.return_value = []
        # Act
        result = weighted_random(mocked_process_step)
        # Assert
        mocked_choices.assert_called_once_with(mock_group_users)
        assert result == user_b

    def test_weighted_random_single_user(self, mocker: MockerFixture, mocked_process_step, mocked_group):
        # Arrange
        user_a = mocker.MagicMock(spec=User, pk=1)
        mocked_group.user_set.all.return_value = [user_a]
        mocked_group.user_set.count.return_value = 1
        mocked_process_step.group = mocked_group
        mocked_process_step.process = mocker.MagicMock(spec=Process)

        mocked_qs = mocker.patch("wbcore.contrib.workflow.workflows.assignees.ProcessStep.objects.filter")

        mocked_query = mocked_qs.return_value
        mocked_query.exclude.return_value.values_list.return_value = []
        # Act
        result = weighted_random(mocked_process_step)
        # Assert
        assert result == user_a

    def test_random_group_member_multiple_users(self, mocker: MockerFixture, mocked_process_step, mocked_group):
        # Arrange
        mocked_process_step = mocker.MagicMock(spec=ProcessStep)

        user_a = mocker.MagicMock(spec=User)
        user_b = mocker.MagicMock(spec=User)
        user_c = mocker.MagicMock(spec=User)
        users = [user_a, user_b, user_c]

        mocked_group.user_set.exists.return_value = True
        mocked_group.user_set.count.return_value = 3
        mocked_group.user_set.all.return_value = users

        mocked_process_step.group = mocked_group

        mock_randint = mocker.patch("wbcore.contrib.workflow.workflows.assignees.randint", return_value=1)
        # Act
        result = random_group_member(mocked_process_step)
        # Assert
        mock_randint.assert_called_once_with(0, 2)
        assert result == user_b  # index 1

    def test_random_group_member_single_user(self, mocker: MockerFixture, mocked_process_step, mocked_group):
        # Arrange
        mocked_process_step = mocker.MagicMock(spec=ProcessStep)

        user_a = mocker.MagicMock(spec=User)
        mocked_group.user_set.exists.return_value = True
        mocked_group.user_set.count.return_value = 1
        mocked_group.user_set.all.return_value = [user_a]

        mocked_process_step.group = mocked_group
        # Act
        result = random_group_member(mocked_process_step)
        # Assert
        assert result == user_a

    def test_random_group_member_empty_group(self, mocker, mocked_process_step, mocked_group):
        # Arrange
        mocked_process_step.group = mocked_group
        mocked_process_step.group.user_set.exists.return_value = False

        mocked_set_failed = mocker.patch.object(mocked_process_step.step.get_casted_step(), "set_failed")
        # Act
        result = random_group_member(mocked_process_step)
        # Assert
        mocked_set_failed.assert_called_once()
        assert result is None

    def test_random_group_member_no_group(self, mocker, mocked_process_step):
        # Arrange
        mocked_process_step.group = None
        mocked_set_failed = mocker.patch.object(mocked_process_step.step.get_casted_step(), "set_failed")
        # Act
        result = random_group_member(mocked_process_step)
        # Assert
        mocked_set_failed.assert_called_once()
        assert result is None



from .formatting import Condition, Formatting, FormattingRule
from .instance_display import *
from .list_display import Calendar, Field, Legend, LegendItem, ListDisplay, Tooltip, TreeGroupLevelOption
from .view_config import DisplayIdentifierViewConfig, DisplayViewConfig



import pytest

from wbcore.contrib.dataloader.tests.test.models import EntityTest


@pytest.mark.django_db
class TestEntities:
    def test_entity_factory(self, entity_test):
        assert entity_test.id, "The Test Entity should have an id if the factory created it"

    def test_entity_creation(self):
        entity_test = EntityTest.objects.create(name="test", dl_parameters={"data": {"path": "abc", "parameters": {}}})
        assert entity_test.id, "It should be possible to create an EntityTest model"



from django.apps import AppConfig
from django.db.models.signals import post_migrate


class AuthenticationConfig(AppConfig):
    name = "wbcore.contrib.authentication"

    def ready(self):
        from wbcore.contrib.authentication.management import initialize_task

        post_migrate.connect(
            initialize_task,
            dispatch_uid="wbcore.contrib.authentication.initialize_task",
        )



from django.utils.translation import gettext as _

from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

POSITION_MENUITEM = MenuItem(
    label=_("Company Positions"),
    endpoint="wbcore:directory:position-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["directory.view_position"],
    ),
    add=MenuItem(
        label=_("Create Position"),
        endpoint="wbcore:directory:position-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user),
            permissions=["directory.add_position"],
        ),
    ),
)

CUSTOMERSTATUS_MENUITEM = MenuItem(
    label=_("Customer Statuses"),
    endpoint="wbcore:directory:customerstatus-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["directory.view_customerstatus"],
    ),
    add=MenuItem(
        label=_("Create Customer Status"),
        endpoint="wbcore:directory:customerstatus-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user),
            permissions=["directory.add_customerstatus"],
        ),
    ),
)

COMPANYTYPE_MENUITEM = MenuItem(
    label=_("Company Types"),
    endpoint="wbcore:directory:companytype-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["directory.view_companytype"],
    ),
    add=MenuItem(
        label=_("Create Company Type"),
        endpoint="wbcore:directory:companytype-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user),
            permissions=["directory.add_companytype"],
        ),
    ),
)


SPECIALIZATION_MENUITEM = MenuItem(
    label=_("Specializations"),
    endpoint="wbcore:directory:specialization-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["directory.view_specialization"],
    ),
    add=MenuItem(
        label=_("Create Specialization"),
        endpoint="wbcore:directory:specialization-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user),
            permissions=["directory.add_specialization"],
        ),
    ),
)



from rest_framework.request import Request
from rest_framework.reverse import reverse

from wbcore.configs.decorators import register_config


@register_config
def notification_config(request: Request) -> tuple[str, dict[str, str]]:
    return "notifications", {
        "endpoint": reverse("wbcore:notifications:notification-list", request=request),
        "token": reverse("wbcore:notifications:token", request=request),
        "unread_notifications": reverse("wbcore:notifications:notification-unread-count", request=request),
    }



from django.utils.translation import gettext as _
from rest_framework.reverse import reverse

from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons import ButtonViewConfig


class StepButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self) -> set:
        if workflow_id := self.view.kwargs.get("workflow_id"):
            return {
                bt.DropDownButton(
                    label=_("New Step"),
                    icon=WBIcon.UNFOLD.icon,
                    buttons=(
                        bt.WidgetButton(
                            endpoint=f"{reverse('wbcore:workflow:startstep-list', args=[], request=self.request)}?workflow={workflow_id}",
                            label=_("New Start Step"),
                            icon=WBIcon.START.icon,
                            new_mode=True,
                        ),
                        bt.WidgetButton(
                            endpoint=f"{reverse('wbcore:workflow:userstep-list', args=[], request=self.request)}?workflow={workflow_id}",
                            label=_("New User Step"),
                            icon=WBIcon.PERSON.icon,
                            new_mode=True,
                        ),
                        bt.WidgetButton(
                            endpoint=f"{reverse('wbcore:workflow:decisionstep-list', args=[], request=self.request)}?workflow={workflow_id}",
                            label=_("New Decision Step"),
                            icon=WBIcon.DECISION_STEP.icon,
                            new_mode=True,
                        ),
                        bt.WidgetButton(
                            endpoint=f"{reverse('wbcore:workflow:splitstep-list', args=[], request=self.request)}?workflow={workflow_id}",
                            label=_("New Split Step"),
                            icon=WBIcon.SPLIT.icon,
                            new_mode=True,
                        ),
                        bt.WidgetButton(
                            endpoint=f"{reverse('wbcore:workflow:joinstep-list', args=[], request=self.request)}?workflow={workflow_id}",
                            label=_("New Join Step"),
                            icon=WBIcon.JOIN_STEP.icon,
                            new_mode=True,
                        ),
                        bt.WidgetButton(
                            endpoint=f"{reverse('wbcore:workflow:scriptstep-list', args=[], request=self.request)}?workflow={workflow_id}",
                            label=_("New Script Step"),
                            icon=WBIcon.SCRIPT.icon,
                            new_mode=True,
                        ),
                        bt.WidgetButton(
                            endpoint=f"{reverse('wbcore:workflow:emailstep-list', args=[], request=self.request)}?workflow={workflow_id}",
                            label=_("New Email Step"),
                            icon=WBIcon.MAIL.icon,
                            new_mode=True,
                        ),
                        bt.WidgetButton(
                            endpoint=f"{reverse('wbcore:workflow:finishstep-list', args=[], request=self.request)}?workflow={workflow_id}",
                            label=_("New Finish Step"),
                            icon=WBIcon.END.icon,
                            new_mode=True,
                        ),
                    ),
                )
            }
        return set()



from .base import WBCoreViewConfig


class DocumentationViewConfig(WBCoreViewConfig):
    metadata_key = "documentation"
    config_class_attribute = "documentation_config_class"

    def get_metadata(self) -> str | None:
        if url := self.view._get_documentation_url(self.instance):  # type: ignore
            return url
        return None



from configurations import values
from corsheaders.defaults import default_headers


class Network:
    ALLOWED_HOSTS = values.ListValue(["*"], environ_prefix=None)
    CORS_ALLOW_ALL_ORIGINS = values.BooleanValue(True, environ_prefix=None)
    CORS_ALLOWED_ORIGINS = values.ListValue([], environ_prefix=None)
    CORS_ALLOW_HEADERS = list(default_headers) + ["WB-DISPLAY-IDENTIFIER"]
    DATA_UPLOAD_MAX_MEMORY_SIZE = values.IntegerValue(2621440, environ_prefix=None)
    DATA_UPLOAD_MAX_NUMBER_FIELDS = values.IntegerValue(1000, environ_prefix=None)
    DATA_UPLOAD_MAX_NUMBER_FILES = values.IntegerValue(100, environ_prefix=None)


class SSLNetwork(Network):
    SECURE_SSL_REDIRECT = True
    SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")
    SESSION_COOKIE_SECURE = True
    CSRF_COOKIE_SECURE = True
